from __future__ import annotations

import abc as _abc
import contextlib as _cl
import datetime as _dt
import email.utils as _eml
import functools as _ft
import hashlib as _hl
import os
import pathlib as _pl
import tempfile as _tf
import time as _t
import typing
import typing as _tp
import urllib.parse as _up
import urllib.request as _ur

import certifi as _cert
import nrrd as _nrrd
import requests as _rq

from .. import config
from .._util import obj_str_insert
from ..config import types
from ..config._attrs import cfglist
from ..morphologies.parsers import MorphologyParser
from ..reporting import warn

if _tp.TYPE_CHECKING:
    from ..core import Scaffold
    from ..morphologies import Morphology
    from ..storage.interfaces import FileStore


def _is_uri(url):
    return _up.urlparse(url).scheme != ""


def _uri_to_path(uri):
    parsed = _up.urlparse(uri)
    host = f"{os.path.sep}{os.path.sep}{parsed.netloc}{os.path.sep}"
    return os.path.relpath(
        os.path.normpath(os.path.join(host, _ur.url2pathname(_up.unquote(parsed.path))))
    )


class FileDependency:
    def __init__(
        self,
        source: str | os.PathLike,
        file_store: FileStore = None,
        ext: str = None,
        cache=True,
    ):
        self._given_source: str = str(source)
        if _is_uri(self._given_source):
            self._uri = self._given_source
        else:
            path = _pl.Path(source).absolute()
            self._uri: str = path.as_uri()
            ext = ext or path.suffix[1:] or None
        self._scheme: FileScheme = _get_scheme(_up.urlparse(self._uri).scheme)
        self.file_store = file_store
        self.extension = ext
        self.cache = cache

    @property
    def uri(self):
        return self._uri

    def __hash__(self):
        return hash(self._uri)

    def __inv__(self):
        return self._given_source

    @obj_str_insert
    def __str__(self):
        return f"'{self._uri}'"

    def get_content(self, check_store=True):
        if check_store:
            stored = self.get_stored_file()
            if stored is None or self._scheme.should_update(self, stored):
                content = self._get_content()
                if self.cache:
                    self.store_content(content, meta=self._scheme.get_meta(self))
                return content
            else:
                return stored.load()
        else:
            return self._get_content()

    def get_meta(self, check_store=True):
        if (
            not check_store or ((stored := self.get_stored_file()) is None)
        ) or self._scheme.should_update(self, stored):
            return self._scheme.get_meta(self)
        else:
            return stored.meta

    @_cl.contextmanager
    def provide_locally(self):
        try:
            path = self._scheme.get_local_path(self)
            if os.path.exists(path):
                yield (path, None)
            else:
                raise FileNotFoundError()
        except (TypeError, FileNotFoundError):
            if self.file_store is None:
                raise FileNotFoundError(f"Can't find {self}") from None
            content = self.get_content()
            with _tf.TemporaryDirectory() as dirpath:
                name = "file"
                if self.extension:
                    name = f"{name}.{self.extension}"
                filepath = os.path.join(dirpath, name)
                with open(filepath, "wb") as f:
                    f.write(content[0])
                yield (filepath, content[1])

    def provide_stream(self):
        return self._scheme.provide_stream(self)

    def get_stored_file(self):
        if not self.file_store:
            raise ValueError(
                "Can't check for file dependency in store before scaffold is ready."
            )
        return self.file_store.find_meta("source", self._given_source)

    def store_content(self, content, encoding=None, meta=None):
        if not self.file_store:
            raise ValueError(
                "Can't store file dependency in store before scaffold is ready."
            )
        if isinstance(content, tuple):
            content, encoding = content
        # Save the file under the same id if it already exists
        id_ = getattr(self.get_stored_file(), "id", None)
        if meta is None:
            meta = {}
        meta.update(self._scheme.get_meta(self))
        meta["source"] = self._given_source
        return self.file_store.store(
            content, meta=meta, encoding=encoding, id=id_, overwrite=True
        )

    def should_update(self):
        if not self.file_store:
            raise ValueError(
                "Can't update file dependency in store before scaffold is ready."
            )
        stored = self.get_stored_file()
        return stored is None or self._scheme.should_update(self, stored)

    def _get_content(self):
        if not self._scheme.find(self):
            raise FileNotFoundError(f"Couldn't find {self._uri}")
        return self._scheme.get_content(self)

    def update(self, force=False):
        if force or self.should_update():
            self.get_content()


class UriScheme(_abc.ABC):
    @_abc.abstractmethod
    def find(self, file: FileDependency):
        path = _uri_to_path(file.uri)
        return os.path.exists(path)

    @_abc.abstractmethod
    def should_update(self, file: FileDependency, stored_file):
        path = _uri_to_path(file.uri)
        try:
            file_mtime = os.path.getmtime(path)
        except FileNotFoundError:
            return False
        try:
            stored_mtime = stored_file.mtime
        except Exception:
            return True
        return file_mtime > stored_mtime

    @_abc.abstractmethod
    def get_content(self, file: FileDependency):
        with self.provide_stream(file) as (fp, encoding):
            fp: _tp.BinaryIO
            return (fp.read(), encoding)

    @_cl.contextmanager
    @_abc.abstractmethod
    def provide_stream(self, file):
        path = _uri_to_path(file.uri)
        with open(path, "rb") as fp:
            yield (fp, None)

    @_abc.abstractmethod
    def get_meta(self, file: FileDependency):
        return {}

    @_abc.abstractmethod
    def get_local_path(self, file: FileDependency):
        raise RuntimeError(f"{type(self)} has no local path representation")


class FileScheme(UriScheme):
    def find(self, file: FileDependency):
        return super().find(file)

    def should_update(self, file: FileDependency, stored_file):
        return super().should_update(file, stored_file)

    def get_content(self, file: FileDependency):
        return super().get_content(file)

    def provide_stream(self, file: FileDependency):
        return super().provide_stream(file)

    def get_meta(self, file: FileDependency):
        return super().get_meta(file)

    def get_local_path(self, file: FileDependency):
        return _uri_to_path(file.uri)


class UrlScheme(UriScheme):
    def resolve_uri(self, file: FileDependency):
        return file.uri

    def find(self, file: FileDependency):
        with self.create_session() as session:
            response = session.head(self.resolve_uri(file), verify=_cert.where())
        return response.status_code == 200

    def should_update(self, file: FileDependency, stored_file):
        mtime = stored_file.mtime
        headers = self.get_meta(file).get("headers", {})
        stored_headers = stored_file.meta.get("headers", {})
        if "ETag" in headers:
            new_etag = headers["ETag"]
            old_etag = stored_headers.get("ETag")
            # Check if we have the latest ETag
            return old_etag != new_etag
        elif "Last-Modified" in headers:
            their_mtime = _dt.datetime(
                *_eml.parsedate(headers["Last-Modified"])[:6]
            ).timestamp()
            return their_mtime > mtime
        # 100h default expiration
        return _t.time() > mtime + 360000

    def get_content(self, file: FileDependency):
        with self.create_session() as session:
            response = session.get(self.resolve_uri(file), verify=_cert.where())
        return (response.content, response.encoding)

    def get_meta(self, file: FileDependency):
        with self.create_session() as session:
            response = session.head(self.resolve_uri(file), verify=_cert.where())
        return {"headers": dict(response.headers)}

    def get_local_path(self, file: FileDependency):
        raise TypeError("URL schemes don't have a local path")

    @_cl.contextmanager
    def provide_stream(self, file):
        with self.create_session() as session:
            response = session.get(
                self.resolve_uri(file), stream=True, verify=_cert.where()
            )
        response.raw.decode_content = True
        response.raw.auto_close = False
        yield (response.raw, response.encoding)

    def create_session(self):
        return _rq.Session()

    def get_base_url(self):
        raise NotImplementedError("Base UrlScheme has no fixed base URL.")


class NeuroMorphoScheme(UrlScheme):
    _nm_url = "https://neuromorpho.org/"
    _meta = "api/neuron/name/"
    _files = "dableFiles/"

    def resolve_uri(self, file: FileDependency):
        meta = self.get_nm_meta(file)
        return self._swc_url(meta["archive"], meta["neuron_name"])

    @_ft.cache
    def get_nm_meta(self, file: FileDependency):
        name = _up.urlparse(file.uri).hostname
        # urlparse gives lowercase, so slice out the original cased NM name
        idx = file.uri.lower().find(name)
        name = file.uri[idx : (idx + len(name))]
        with self.create_session() as session:
            try:
                res = session.get(self._nm_url + self._meta + name, verify=_cert.where())
            except Exception:
                return {"archive": "none", "neuron_name": "none"}
            if res.status_code == 404:
                res = session.get(self._nm_url, verify=_cert.where())
                if res.status_code != 200 or "Service Interruption Notice" in res.text:
                    warn(f"NeuroMorpho.org is down, can't retrieve morphology '{name}'.")
                    return {"archive": "none", "neuron_name": "none"}
                raise OSError(f"'{name}' is not a valid NeuroMorpho name.")
            elif res.status_code != 200:
                raise OSError("NeuroMorpho API error: " + res.text)
        return res.json()

    def get_base_url(self):
        return self._nm_url

    def get_meta(self, file: FileDependency):
        meta = super().get_meta(file)
        meta["neuromorpho_data"] = self.get_nm_meta(file)
        return meta

    @classmethod
    def _swc_url(cls, archive, name):
        base_url = f"{cls._nm_url}{cls._files}{_up.quote(archive.lower())}"
        return f"{base_url}/CNG%20version/{name}.CNG.swc"

    def create_session(self):
        # Weak DH key on neuromorpho.org
        # https://stackoverflow.com/a/76217135/1016004
        from requests.adapters import HTTPAdapter
        from urllib3 import PoolManager
        from urllib3.util import create_urllib3_context

        class DHAdapter(HTTPAdapter):
            def init_poolmanager(self, connections, maxsize, block=False, **kwargs):
                ctx = create_urllib3_context(ciphers=":HIGH:!DH:!aNULL")
                self.poolmanager = PoolManager(
                    num_pools=connections,
                    maxsize=maxsize,
                    block=block,
                    ssl_context=ctx,
                    **kwargs,
                )

        session = _rq.Session()
        session.mount(self._nm_url, DHAdapter())
        return session


@_ft.cache
def _get_schemes() -> _tp.Mapping[str, type[FileScheme]]:
    from ..plugins import discover

    schemes = discover("storage.schemes")
    schemes["file"] = FileScheme
    schemes["http"] = schemes["https"] = UrlScheme
    schemes["nm"] = NeuroMorphoScheme
    return schemes


def _get_scheme(scheme: str) -> FileScheme:
    schemes = _get_schemes()
    try:
        return schemes[scheme]()
    except KeyError:
        raise KeyError(f"{scheme} is not a known file scheme.") from None


@config.node
class FileDependencyNode:
    scaffold: Scaffold
    file: FileDependency = config.attr(type=FileDependency)

    def __init__(self, value=None, **kwargs):
        if value is not None:
            self.file = value

    def __boot__(self):
        self.file.file_store = self.scaffold.files
        if self.scaffold.is_main_process():
            self.file.update()

    def __inv__(self):
        if not isinstance(self, FileDependencyNode):
            return self
        if self._config_pos_init:
            return self.file._given_source
        else:
            tree = self.__tree__()
            return tree

    def load_object(self):
        return self.file.get_content()

    def provide_locally(self):
        return self.file.provide_locally()

    def provide_stream(self):
        return self.file.provide_stream()

    def get_stored_file(self):
        return self.file.get_stored_file()


@config.node
class CodeDependencyNode(FileDependencyNode):
    """
    Allow the loading of external code during network loading.
    """

    module: str = config.attr(type=str, required=types.shortform())
    """
    Should be either the path to a python file or a import like string.
    """
    attr: str = config.attr(type=str)
    """
    Attribute to extract from the loaded script.
    """

    @config.property
    def file(self):
        if getattr(self, "scaffold", None) is not None:
            file_store = self.scaffold.files
        else:
            file_store = None
        if os.path.isfile(self.module):
            # Convert potential relative path to absolute path
            module_file = os.path.abspath(os.path.join(os.getcwd(), self.module))
        else:
            # Module like string converted to a path string relative to current folder
            module_file = "./" + self.module.replace(".", os.sep) + ".py"
        return FileDependency(module_file, file_store=file_store)

    def __init__(self, module=None, **kwargs):
        super().__init__(**kwargs)
        if module is not None:
            self.module = module

    def __inv__(self):
        if not isinstance(self, CodeDependencyNode):
            return self
        res = {"module": self.module}
        if self.attr is not None:
            res["attr"] = self.attr
        return res

    def load_object(self):
        import importlib.util
        import sys

        sys.path.append(os.getcwd())
        try:
            with self.file.provide_locally() as (path, encoding):
                spec = importlib.util.spec_from_file_location(self.module, path)
                module = importlib.util.module_from_spec(spec)
                sys.modules[self.module] = module
                spec.loader.exec_module(module)
                return module if self.attr is None else getattr(module, self.attr)
        finally:
            tmp = list(reversed(sys.path))
            tmp.remove(os.getcwd())
            sys.path = list(reversed(tmp))


class OperationCallable(typing.Protocol):
    def __call__(self, obj: object, **kwargs: typing.Any) -> object:
        pass


@config.node
class Operation:
    func: OperationCallable = config.attr(type=types.function_())
    parameters: dict[typing.Any] = config.catch_all(type=types.any_())

    def __init__(self, value=None, /, **kwargs):
        if value is not None:
            self.func = value

    def __call__(self, obj):
        return self.func(obj, **self.parameters)


class FilePipelineMixin:
    pipeline: cfglist[Operation] = config.list(type=Operation)

    def pipe(self, input):
        return _ft.reduce(lambda state, func: func(state), self.pipeline, input)


@config.node
class NrrdDependencyNode(FilePipelineMixin, FileDependencyNode):
    """
    Configuration dependency node to load NRRD files.
    """

    def get_header(self):
        with self.file.provide_locally() as (path, encoding):
            return _nrrd.read_header(path)

    def get_data(self):
        with self.file.provide_locally() as (path, encoding):
            return _nrrd.read(path)[0]

    def load_object(self):
        return self.pipe(self.get_data())


class MorphologyOperationCallable(OperationCallable):
    """
    Hello.
    """

    def __call__(self, obj: Morphology, **kwargs: typing.Any) -> Morphology:
        pass


@config.node
class MorphologyOperation(Operation):
    func: MorphologyOperationCallable = config.attr(
        type=types.method_shortcut("bsb.morphologies.Morphology")
    )


@config.node
class MorphologyDependencyNode(FilePipelineMixin, FileDependencyNode):
    """
    Configuration dependency node to load morphology files.

    The content of these files will be stored in bsb.morphologies.Morphology instances.
    """

    pipeline: cfglist[MorphologyOperation] = config.list(type=MorphologyOperation)
    name: str = config.attr(type=str, default=None, required=False)
    parser: MorphologyParser = config.attr(type=MorphologyParser, default={})
    """
    Name associated to the morphology.

    If not provided, the program will use the name of the file in which the morphology is
    stored.
    """

    def store_content(self, content, *args, encoding=None, meta=None):
        if meta is None:
            meta = {}
        meta["_hash"] = self._hash(content)
        meta["_stale"] = True
        stored = super().store_content(content, *args, encoding=encoding, meta=meta)
        return stored

    def load_object(self, parser=None, save=True) -> Morphology:
        if parser is None or self.__class__.parser.is_dirty(self):
            parser = self.parser
        self.file.update()
        stored = self.get_stored_file()
        meta = stored.meta
        if meta.get("_stale", True):
            content, meta = stored.load()
            if hasattr(parser, "parse_content"):
                morpho = parser.parse_content(
                    content.decode(meta.get("encoding", "utf8"))
                )
            else:
                morpho = parser.parse(self.file)
            morpho.meta = meta
            morpho = self.pipe(morpho)
            meta["hash"] = self._hash(content)
            meta["_stale"] = False
            morpho.meta = meta
            if save:
                self.scaffold.morphologies.save(
                    self.get_morphology_name(), morpho, overwrite=True
                )
            return morpho
        else:
            return self.scaffold.morphologies.load(self.get_morphology_name())

    def get_morphology_name(self):
        """
        Returns morphology name provided by the user or extract it from its file name.

        :returns: Morphology name
        :rtype: str
        """
        return self.name if self.name is not None else _pl.Path(self.file.uri).stem

    def store_object(self, morpho, hash_):
        """
        Save a morphology into the circuit file under the name of this instance
        morphology.

        :param hash_: Hash key to store as metadata with the morphology
        :type hash_: str
        :param morpho: Morphology to store
        :type morpho: bsb.morphologies.Morphology
        """
        self.scaffold.morphologies.save(
            self.get_morphology_name(), morpho, meta={"hash": hash_}
        )

    def _hash(self, content):
        md5 = _hl.new("md5", usedforsecurity=False)
        if isinstance(content, str):
            md5.update(content.encode("utf-8"))
        else:
            md5.update(content)
        return md5.hexdigest()

    def queue(self, pool):
        """
        Add the loading of the current morphology to a job queue.

        :param pool: Queue of jobs.
        :type pool: bsb.services.pool.JobPool
        """

        def create_morphology(scaffold, i):
            scaffold.configuration.morphologies[i].load_object()

        pool.queue(
            create_morphology, (self._config_index,), submitter=self, uri=self.file.uri
        )


@config.node
class MorphologyPipelineNode(FilePipelineMixin):
    files: list[MorphologyDependencyNode] = config.list(
        type=MorphologyDependencyNode, required=True
    )
    pipeline: cfglist[MorphologyOperation] = config.list(type=MorphologyOperation)
    parser: MorphologyParser = config.attr(type=MorphologyParser, required=False)

    def queue(self, pool):
        """
        Add the loading of the current morphology to a job queue.

        :param pool: Queue of jobs.
        :type pool: bsb.services.pool.JobPool
        """

        def job(scaffold, i, j):
            me = scaffold.configuration.morphologies[i]
            fnode = me.files[j]
            morpho = fnode.load_object(parser=me.parser, save=False)
            morpho_out = me.pipe(morpho)
            scaffold.morphologies.save(
                fnode.get_morphology_name(), morpho_out, overwrite=True
            )

        for k in range(len(self.files)):
            pool.queue(
                job,
                (self._config_index, k),
                submitter=self,
                node=k,
                uri=self.files[k].file.uri,
            )


__all__ = [
    "CodeDependencyNode",
    "FileDependency",
    "FileDependencyNode",
    "FileScheme",
    "MorphologyDependencyNode",
    "MorphologyOperation",
    "NeuroMorphoScheme",
    "NrrdDependencyNode",
    "Operation",
    "UriScheme",
    "UrlScheme",
]



import importlib

from ..exceptions import DependencyError


class ErrorModule:
    def __init__(self, message):
        self._msg = message

    def __getattr__(self, attr):
        raise DependencyError(self._msg)


class MockModule:
    def __new__(cls, module):
        try:
            instance = importlib.import_module(module)
        except ImportError:
            instance = super().__new__(cls)
            instance._mocked = True
        return instance



"""
bsb.config module.

Contains the dynamic attribute system; Use ``@bsb.config.root/node/dynamic/pluggable`` to
decorate your classes and add class attributes using ``x =
config.attr/dict/list/ref/reflist`` to populate your classes with powerful attributes.
"""

import builtins
import contextlib
import functools
import glob
import itertools
import os
import sys
import traceback
import typing
from shutil import copy2 as copy_file

from .. import plugins
from .._util import ichain
from ..exceptions import ConfigTemplateNotFoundError, ParserError
from . import refs, types
from ._attrs import (
    ConfigurationAttribute,
    attr,
    catch_all,
    dict,
    dynamic,
    file,
    list,
    node,
    pluggable,
    property,
    provide,
    ref,
    reflist,
    root,
    slot,
    unset,
)
from ._distributions import Distribution
from ._hooks import after, before, has_hook, on, run_hook
from ._make import (
    compose_nodes,
    get_config_attributes,
    walk_node_attributes,
    walk_nodes,
)
from .parsers import get_configuration_parser, get_configuration_parser_classes

if typing.TYPE_CHECKING:
    from ._config import Configuration


@functools.cache
def __getattr__(name):
    if name == "Configuration":
        # Load the Configuration class on demand, not on import, to avoid circular
        # dependencies.
        from ._config import Configuration

        return Configuration
    else:
        raise object.__getattribute__(sys.modules[__name__], name)


def get_config_path():
    import os

    env_paths = os.environ.get("BSB_CONFIG_PATH", None)
    env_paths = () if env_paths is None else env_paths.split(":")
    plugin_paths = plugins.discover("config.templates")
    return [*itertools.chain((os.getcwd(),), env_paths, *plugin_paths.values())]


def get_configuration_template(template, path=None):
    """
    Returns the configuration template files matching the provided name.

    :param str template: name of the configuration template
    :param list path: list of paths to search for configuration templates
    :rtype: list[str]
    """
    path = [
        *map(
            os.path.abspath,
            itertools.chain(get_config_path(), path or ()),
        )
    ]
    for d in path:
        if files := glob.glob(os.path.join(d, template)):
            break
    else:
        raise ConfigTemplateNotFoundError(
            "'%template%' not found in config path %path%", template, path
        )
    return files


def copy_configuration_template(template, output="network_configuration.json", path=None):
    """
    Copy the first configuration template file matching the provided name to the provided
    output filename.

    :param str template: name of the configuration template
    :param str output: name of the output file
    :param list path: list of paths to search for configuration templates
    """
    copy_file(get_configuration_template(template, path)[0], output)


def format_configuration_content(parser_name: str, config: "Configuration", **kwargs):
    """
    Convert a configuration object to a string using the given parser.
    """
    return get_configuration_parser(parser_name, **kwargs).generate(
        config.__tree__(), pretty=True
    )


def make_configuration_diagram(config):
    dot = f'digraph "{config.name or "network"}" {{'
    for c in config.cell_types.values():
        dot += f'\n  {c.name}[label="{c.name}"]'
    for name, conn in config.connectivity.items():
        for pre in conn.presynaptic.cell_types:
            for post in conn.postsynaptic.cell_types:
                dot += f'\n  {pre.name} -> {post.name}[label="{name}"];'
    dot += "\n}\n"
    return dot


def _try_parsers(content, classes, ext=None, path=None):  # pragma: nocover
    if ext is not None:

        def file_has_parser_ext(kv):
            return ext not in getattr(kv[1], "data_extensions", ())

        classes = builtins.dict(sorted(classes.items(), key=file_has_parser_ext))
    exc = {}
    for name, cls in classes.items():
        try:
            tree, meta = cls().parse(content, path=path)
        except Exception as e:
            if getattr(e, "_bsbparser_show_user", False):
                raise e from None
            exc[name] = e
        else:
            return (name, tree, meta)
    msges = [
        (
            f"- Can't parse contents with '{n}':\n",
            "".join(traceback.format_exception(type(e), e, e.__traceback__)),
        )
        for n, e in exc.items()
    ]
    if path:
        msg = f"Could not parse '{path}'"
    else:
        msg = f"Could not parse content string ({len(content)} characters long)"
    raise ParserError("\n".join(ichain(msges)) + f"\n{msg}")


def _from_parsed(parser_name: str, tree: dict, meta: dict, file: str = None):
    from ._config import Configuration

    conf = Configuration(tree)
    conf._parser = parser_name
    conf._meta = meta
    conf._file = file
    return conf


def parse_configuration_file(file, parser=None, path=None, **kwargs):
    """
    Parse a configuration file into a Configuration.

    :param str file: path to the configuration file
    :param str path: path from which to parse the configuration file. Use this
     parameter to mock the configuration file location which can help resolve
     relative paths in configuration files.
    :param ConfigurationParser parser: parser to use. If None is provided, the BSB will
     parse the content based on the file extension defined in `path` and the available
     ConfigurationParser classes that the BSB could find.
    :return: the parsed configuration.
    :rtype: Configuration
    """
    if hasattr(file, "read"):
        data = file.read()
        with contextlib.suppress(TypeError):
            path = str(path) or os.fspath(file)
    else:
        file = os.path.abspath(file)
        path = path or file
        with open(file) as f:
            data = f.read()
    return parse_configuration_content(data, parser, path, **kwargs)


def parse_configuration_content_to_dict(content, parser=None, path=None, **kwargs):
    """
    Parse the content of a configuration file into a dictionary.

    :param str content: content of a configuration file
    :param str path: path to the location of the configuration file
    :param ConfigurationParser parser: parser to use. If None is provided, the BSB will
     parse the content based on the file extension defined in `path` and the available
     ConfigurationParser classes that the BSB could find.
    :return: a dictionary of the parsed content, a dictionary storing the metadata, and
     the name of the parser used to obtain the results.
    :rtype: (dict, dict, str)
    """
    if parser is None:
        parser_classes = get_configuration_parser_classes()
        ext = path.split(".")[-1] if path is not None else None
        parser_name, tree, meta = _try_parsers(content, parser_classes, ext, path=path)
    elif isinstance(parser, str):
        parser_name = parser
        parser = get_configuration_parser(parser_name, **kwargs)
        tree, meta = parser.parse(content, path=path)
    else:
        parser_name = parser.__name__
        tree, meta = parser.parse(content, path=path)
    return tree, meta, parser_name


def parse_configuration_content(content, parser=None, path=None, **kwargs):
    """
    Parse the content of a configuration file into a Configuration.

    :param str content: content of a configuration file
    :param str path: path to the location of the configuration file
    :param ConfigurationParser parser: parser to use. If None is provided, the BSB will
     parse the content based on the file extension defined in `path` and the available
     ConfigurationParser classes that the BSB could find.
    :return: the parsed configuration.
    :rtype: Configuration
    """
    tree, meta, parser_name = parse_configuration_content_to_dict(
        content, parser, path, **kwargs
    )
    return _from_parsed(parser_name, tree, meta, path)


# Static public API
__all__ = [
    "Configuration",
    "ConfigurationAttribute",
    "Distribution",
    "after",
    "attr",
    "before",
    "catch_all",
    "compose_nodes",
    "copy_configuration_template",
    "dict",
    "dynamic",
    "file",
    "format_configuration_content",
    "get_config_attributes",
    "get_config_path",
    "has_hook",
    "list",
    "make_configuration_diagram",
    "node",
    "on",
    "parse_configuration_file",
    "parse_configuration_content_to_dict",
    "parse_configuration_content",
    "pluggable",
    "property",
    "provide",
    "ref",
    "refs",
    "reflist",
    "root",
    "run_hook",
    "slot",
    "types",
    "unset",
    "walk_node_attributes",
    "walk_nodes",
]
__api__ = [
    "Configuration",
    "ConfigurationAttribute",
    "Distribution",
    "compose_nodes",
    "copy_configuration_template",
    "format_configuration_content",
    "get_config_attributes",
    "get_config_path",
    "make_config_diagram",
    "parse_configuration_file",
    "parse_configuration_content_to_dict",
    "parse_configuration_content",
    "refs",
    "types",
    "walk_node_attributes",
    "walk_nodes",
]



"""
Module for the CellType configuration node and its dependencies.
"""

import typing

from . import config
from ._util import obj_str_insert
from .config import types
from .placement.indicator import PlacementIndications

if typing.TYPE_CHECKING:
    from .core import Scaffold


@config.node
class Plotting:
    scaffold: "Scaffold"

    display_name = config.attr()
    """
    Label used to display this cell type in plots.
    """
    color = config.attr()
    """
    Color used to display this cell type in plots.
    """
    opacity = config.attr(type=types.fraction(), default=1.0)
    """
    Opacity used to display this cell type in plots.
    """


def _not_an_entity(section):
    return "entity" not in section or not bool(section["entity"])


@config.node
class CellType:
    """
    Information on a population of cells.
    """

    scaffold: "Scaffold"

    name = config.attr(key=True)
    """
    Name of the cell type, equivalent to the key it occurs under in the configuration.
    """
    spatial = config.attr(
        type=PlacementIndications, required=_not_an_entity, default={"radius": None}
    )
    """
    Spatial information about the cell type such as radius and density, and geometric or
    morphological information.
    """
    plotting = config.attr(type=Plotting)
    """
    Plotting information about the cell type, such as color and labels.
    """
    entity = config.attr(type=bool, default=False)
    """
    Whether this cell type is an entity type.

    Entity types don't have representations in space, but can still be connected and
    simulated.
    """

    def __boot__(self):
        storage = self.scaffold.storage
        if storage.supports("PlacementSet"):
            storage.require_placement_set(self)

    def __lt__(self, other):
        try:
            return self.name < other.name
        except Exception:
            return True

    @obj_str_insert
    def __repr__(self):
        try:
            placements = len(self.get_placement())
        except Exception:
            placements = "?"
        try:
            cells_placed = len(self.get_placement_set())
        except Exception:
            cells_placed = 0
        return f"'{self.name}', {cells_placed} cells, {placements} placement strategies"

    def get_placement(self):
        """
        Get the placement components this cell type is a part of.
        """
        return self.scaffold.get_placement_of(self)

    def get_placement_set(self, *args, **kwargs):
        """
        Retrieve this cell type's placement data.

        :param chunks: When given, restricts the placement data to these chunks.
        :type chunks: list[bsb.storage._chunks.Chunk]
        """
        return self.scaffold.get_placement_set(self, *args, **kwargs)

    def get_morphologies(self):
        """
        Return the list of morphologies of this cell type.

        :rtype: list[~bsb.storage.interfaces.StoredMorphology]
        """
        if "morphologies" not in self.spatial:
            return []
        else:
            return self.scaffold.storage.morphologies.select(*self.spatial.morphologies)

    def clear(self):
        """
        Clear all the placement and connectivity data associated with this cell type.
        """
        self.clear_placement()
        self.clear_connections()

    def clear_placement(self):
        """
        Clear all the placement data associated with this cell type.

        Connectivity data will remain, but be invalid.
        """
        self.get_placement_set().clear()

    def clear_connections(self):
        """
        Clear all the connectivity data associated with this cell type.

        Any connectivity set that this cell type is a part of will be entirely removed.
        """
        for conn_set in self.scaffold.get_connectivity_sets():
            if self is conn_set.pre_type or self is conn_set.post_type:
                conn_set.clear()

    # This property was mostly added so that accidental assignment to `ct.morphologies`
    # instead of `ct.spatial.morphologies` raises an error.
    @property
    def morphologies(self):
        return self.get_morphologies()

    @morphologies.setter
    def morphologies(self, value):
        raise AttributeError(
            "`cell_type.morphologies` is a readonly attribute. Did you mean"
            " `cell_type.spatial.morphologies`?"
        )


__all__ = ["CellType", "PlacementIndications", "Plotting"]



"""
Module for the Region types.
"""

from __future__ import annotations

import abc
import typing

import numpy as np

from .. import config
from ..config import refs, types
from ..reporting import warn
from ._layout import Layout

if typing.TYPE_CHECKING:
    from ..core import Scaffold
    from .partition import Partition


@config.dynamic(attr_name="type", required=False, default="group", auto_classmap=True)
class Region(abc.ABC):
    """
    Base region.

    When arranging will simply call arrange/layout on its children but won't cause any
    changes itself.
    """

    scaffold: Scaffold

    name: str = config.attr(key=True)
    children: list[Region | Partition] = config.reflist(
        refs.regional_ref, backref="region", required=True
    )
    """
    Reference to Regions or Partitions belonging to this region.
    """

    @property
    def data(self):
        # The data property is read-only to users, but `_data` is assigned
        # during the layout process
        return self._data

    def get_dependencies(self):
        return self.children.copy()

    def get_layout(self, hint):
        layouts = [dep.get_layout(hint) for dep in self.get_dependencies()]
        return Layout(hint.data.copy(), owner=self, children=layouts)

    def do_layout(self, hint):
        layout = self.get_layout(hint)
        layout.accept()

    @abc.abstractmethod
    def rotate(self, rotation):  # pragma: nocover
        pass

    @abc.abstractmethod
    def translate(self, offset):  # pragma: nocover
        pass

    @abc.abstractmethod
    def scale(self, factors):  # pragma: nocover
        pass


@config.node
class RegionGroup(Region, classmap_entry="group"):
    """
    Base implementation of Region.

    Any transformation on the region will be directly applied to its children (Regions or
    Partitions).
    """

    def rotate(self, rotation):
        for child in self.children:
            child.rotate(rotation)

    def translate(self, offset):
        for child in self.children:
            child.translate(offset)

    def scale(self, factors):
        for child in self.children:
            child.scale(factors)


@config.node
class Stack(RegionGroup, classmap_entry="stack"):
    """
    Stack components on top of each other and adjust its own height accordingly.
    """

    axis: typing.Literal["x"] | typing.Literal["y"] | typing.Literal["z"] = config.attr(
        type=types.in_(["x", "y", "z"]), default="z"
    )
    """
    Axis along which the stack's children will be stacked.
    """
    anchor: Region | Partition = config.ref(refs.regional_ref)
    """
    Reference to one child of the stack, which origin will become the origin of the stack.
    """

    def _resolve_anchor_offset(self, children, axis_idx):
        """
        Check if the anchor is one of the children of the stack and if so, return the
        offset so the anchor is at the origin of the stack.
        """
        children_owners = [child._owner for child in children]
        if self.anchor is not None and self.anchor in children_owners:
            index = children_owners.index(self.anchor)
            return children[index].data.ldc[axis_idx] - sum(
                children[i].data.dimensions[axis_idx] for i in range(index)
            )
        else:
            # if anchor is not defined or one of the children
            # then the origin of the stack corresponds to the origin of the first child
            return children[0].data.ldc[axis_idx]

    def get_layout(self, hint):
        layout = super().get_layout(hint)
        axis_idx = ("x", "y", "z").index(self.axis)
        trans_eye = np.zeros(3)
        trans_eye[axis_idx] = 1

        cumul_offset = self._resolve_anchor_offset(layout.children, axis_idx)
        for child in layout.children:
            if child.data is None:
                warn(f"Skipped layout arrangement of {child._owner.name} in {self.name}")
                continue
            translation = (
                layout.data.ldc[axis_idx] + cumul_offset - child.data.ldc
            ) * trans_eye
            if not np.allclose(0, translation):
                child.propose_translate(translation)
            cumul_offset += child.data.dimensions[axis_idx]
        ldc = layout.data.ldc
        mdc = layout.data.mdc
        mdc[axis_idx] = ldc[axis_idx] + cumul_offset
        return layout

    def rotate(self, rotation):
        for child in self.children:
            child.rotate(rotation)

    def translate(self, offset):
        for child in self.children:
            child.translate(offset)

    def scale(self, factors):
        for child in self.children:
            child.scale(factors)


__all__ = ["Region", "RegionGroup", "Stack"]



import functools
from inspect import signature as _inspect_signature

_reserved_keywords = ["_parent", "_key"]


def _wrap_reserved(t):
    """
    Wrap a type handler in a wrapper that accepts all reserved keyword arguments that the
    config system will push into the type handler call, and pass only those that the
    original type handler accepts.

    This way type handlers can accept any combination of the reserved keyword args without
    raising TypeErrors when they do not accept one.
    """
    from .types import TypeHandler

    # Type handlers never need to be wrapped. The `__init_subclass__` of the TypeHandler
    # class handles wrapping of `__call__` implementations so that they accept and strip
    # _parent & _key.
    if isinstance(t, TypeHandler):
        return t

    # Check which reserved keywords the function already takes
    passes = _reserved_kw_passes(t)
    if all(passes.values()):
        return t

    # Create the keyword arguments of the outer function that accepts all reserved kwargs
    reserved_keys = "".join(f", {key}=None" for key in _reserved_keywords)
    header = f"def type_handler(value, *args{reserved_keys}, **kwargs):\n"
    passes = "".join(f", {key}={key}" for key in _reserved_keywords if passes[key])
    # Create the call to the inner function that is passed only the kwargs that it accepts
    wrap = f" return orig(value, *args{passes}, **kwargs)"
    # Compile the code block and indicate that the function was compiled here.
    mod = compile(header + wrap, f"{__file__}/<_wrap_reserved:compile>", "exec")
    # Execute the code block in this local scope and pick the function out of the scope
    exec(mod, {"orig": t}, bait := locals())
    type_handler = bait["type_handler"]
    # Copy over the metadata of the original function
    type_handler = functools.wraps(t)(type_handler)
    type_handler.__name__ = t.__name__
    return type_handler


def _reserved_kw_passes(f):
    # Inspect the signature and wrap the typecast in a wrapper that will accept and
    # strip the missing 'key' kwarg
    try:
        sig = _inspect_signature(f)
        params = sig.parameters
    except (ValueError, TypeError):
        params = []

    return {key: key in params for key in _reserved_keywords}



import typing

import numpy as np

from .. import config
from ..config import refs, types
from ..config._attrs import cfglist
from ..exceptions import IndicatorError, PlacementError, PlacementRelationError
from ..morphologies.selector import MorphologySelector

if typing.TYPE_CHECKING:
    from ..cell_types import CellType
    from ..core import Scaffold


@config.node
class PlacementIndications:
    scaffold: "Scaffold"
    radius: float = config.attr(type=float)
    density: float = config.attr(type=float)
    planar_density: float = config.attr(type=float)
    count_ratio: float = config.attr(type=float)
    local_count_ratio: float = config.attr(type=float)
    density_ratio: float = config.attr(type=float)
    relative_to: "CellType" = config.ref(refs.cell_type_ref)
    count: int = config.attr(type=int)
    geometry: dict = config.dict(type=types.any_())
    morphologies: cfglist[MorphologySelector] = config.list(type=MorphologySelector)
    density_key: str = config.attr(type=str)


class _Noner:
    def __getattr__(self, attr):
        return None


class PlacementIndicator:
    def __init__(self, strat, cell_type):
        self._strat = strat
        self._cell_type = cell_type

    @property
    def cell_type(self):
        return self._cell_type

    def get_radius(self):
        return self.assert_indication("radius")

    @property
    def partitions(self):
        return self._strat.partitions

    def use_morphologies(self):
        return bool(self.indication("morphologies"))

    def indication(self, attr):
        ind_strat = self._strat.overrides.get(self._cell_type.name) or _Noner()
        ind_ct = self._cell_type.spatial
        strat = getattr(ind_strat, attr)
        ct = getattr(ind_ct, attr)
        if strat is not None:
            return strat
        return ct

    def assert_indication(self, attr):
        ind = self.indication(attr)
        if ind is None:
            raise IndicatorError(
                f"No configuration indicators found for the {attr}"
                f" of '{self._cell_type.name}' in '{self._strat.name}'"
            )
        return ind

    def guess(self, chunk=None, voxels=None):
        """
        Estimate the count of cell to place based on the cell_type's PlacementIndications.
        Float estimates are converted to int using an acceptance- rejection method.

        :param chunk: if provided, will estimate the number of cell within the Chunk.
        :type chunk: bsb.storage._chunks.Chunk
        :param voxels: if provided, will estimate the number of cell within the VoxelSet.
            Only for cells with the indication "density_key" set or with the indication
            "relative_to" set and the target cell has the indication "density_key" set.
        :type voxels: bsb.voxels.VoxelSet
        :returns: Cell counts for each chunk or voxel.
        :rtype: numpy.ndarray[int]
        """
        count = self.indication("count")
        density = self.indication("density")
        density_key = self.indication("density_key")
        planar_density = self.indication("planar_density")
        relative_to = self.indication("relative_to")
        density_ratio = self.indication("density_ratio")
        count_ratio = self.indication("count_ratio")
        local_count_ratio = self.indication("local_count_ratio")
        if count is not None:
            estimate = self._estim_for_chunk(chunk, count)
        if density is not None:
            estimate = self._density_to_estim(density, chunk)
        if planar_density is not None:
            estimate = self._pdensity_to_estim(planar_density, chunk)
        if relative_to is not None:
            relation = relative_to
            if count_ratio is not None:
                # The total counts of cell of the current strategy is a ratio
                # of the total number of cell placed by the target strategy.
                # This number is uniformly distributed across the current
                # strategy's partition(s).
                strats = self._strat.scaffold.get_placement_of(relation)
                estimate = self._estim_for_chunk(
                    chunk,
                    sum(PlacementIndicator(s, relation).guess() for s in strats)
                    * count_ratio,
                )
            elif local_count_ratio is not None:
                # This count estimate is the ratio of the number of cell of the
                # target strategy that were placed in the current chunk.
                strats = self._strat.scaffold.get_placement_of(relation)
                estimate = (
                    sum(
                        PlacementIndicator(s, relation).guess(chunk, voxels)
                        for s in strats
                    )
                    * local_count_ratio
                )
            elif density_ratio is not None:
                # Create an indicator based on this strategy for the related CT.
                # This means we'll read only the CT indications, and ignore any
                # overrides of other strats, but one can set overrides for the
                # related type in this strat.
                rel_ind = PlacementIndicator(self._strat, relation)
                rel_density = rel_ind.indication("density")
                rel_pl_density = rel_ind.indication("planar_density")
                rel_pl_density_key = rel_ind.indication("density_key")
                if rel_density is not None:
                    estimate = self._density_to_estim(rel_density * density_ratio, chunk)
                elif rel_pl_density is not None:
                    estimate = self._pdensity_to_estim(
                        rel_pl_density * density_ratio, chunk
                    )
                elif rel_pl_density_key is not None:
                    # Use the relation's `guess` to guess according to the relation's
                    # density key
                    estimate = rel_ind.guess(chunk, voxels) * density_ratio
                else:
                    raise PlacementRelationError(
                        f"{self.cell_type.name} requires relation {relation.name}"
                        + " to specify density information."
                    )
            else:
                raise PlacementError(
                    "Relation specified but no ratio indications provided."
                )
        if density_key is not None:
            if voxels is None:
                raise PlacementError("Can't guess voxel density without a voxelset.")
            elif density_key in voxels.data_keys:
                estimate = self._estim_for_voxels(voxels, density_key)
            else:
                raise RuntimeError(
                    f"No voxel density data column '{density_key}' found in any of the"
                    " following partitions:\n"
                    + "\n".join(
                        f"* {p.name}: "
                        + (
                            fstr
                            if (
                                fstr := ", ".join(
                                    f"'{col}'" for col in p.voxelset.data_keys
                                )
                            )
                            else "no data"
                        )
                        for p in self._strat.partitions
                        if hasattr(p, "voxelset")
                    )
                    + "\n".join(
                        f"* {p.name} contains no voxelsets"
                        for p in self._strat.partitions
                        if not hasattr(p, "voxelset")
                    )
                )
        try:
            estimate = np.array(estimate)
        except NameError:
            # If `estimate` is undefined after all this then error out.
            raise IndicatorError(
                "No configuration indicators found for the number of"
                + f"'{self._cell_type.name}' in '{self._strat.name}'"
            ) from None
        if not np.allclose(estimate, estimate // 1):
            # 1.2 cells == 0.8 probability for 1, 0.2 probability for 2
            return (
                np.floor(estimate) + (np.random.rand(estimate.size) < estimate % 1)
            ).astype(int)
        else:
            return np.round(estimate).astype(int)

    def _density_to_estim(self, density, chunk=None):
        return sum(p.volume(chunk) * density for p in self._strat.partitions)

    def _pdensity_to_estim(self, planar_density, chunk=None):
        return sum(p.surface(chunk) * planar_density for p in self._strat.partitions)

    def _estim_for_chunk(self, chunk, count):
        if chunk is None:
            return count
        # When getting with absolute count for a chunk give back the count
        # proportional to the volume in this chunk vs total volume
        chunk_volume = sum(p.volume(chunk) for p in self._strat.partitions)
        total_volume = sum(p.volume() for p in self._strat.partitions)
        return count * chunk_volume / total_volume

    def _estim_for_voxels(self, voxels, key):
        return voxels.get_data(key).ravel().astype(float) * np.prod(
            voxels.get_size_matrix(copy=False), axis=1
        )


__all__ = ["PlacementIndicator"]



from .. import config
from .component import SimulationComponent


@config.node
class DeviceModel(SimulationComponent):
    def implement(self, adapter, simulation, simdata):
        raise NotImplementedError(
            "The "
            + self.__class__.__name__
            + " device does not implement any `implement` function."
        )


__all__ = ["DeviceModel"]



import abc as _abc
import itertools
import typing
from graphlib import TopologicalSorter

from . import _util as _gutil
from .reporting import warn
from .storage._chunks import Chunk

if typing.TYPE_CHECKING:
    from .services import JobPool


def _queue_placement(self, pool: "JobPool", chunk_size):
    # Get the queued jobs of all the strategies we depend on.
    deps = set(
        itertools.chain(*(pool.get_submissions_of(strat) for strat in self.get_deps()))
    )
    # todo: perhaps pass the volume or partition boundaries as chunk size
    pool.queue_placement(self, Chunk([0, 0, 0], None), deps=deps)


def _all_chunks(iter_):
    return _gutil.unique(
        _gutil.ichain(ct.get_placement_set().get_all_chunks() for ct in iter_)
    )


def _queue_connectivity(self, pool: "JobPool"):
    """
    Get the queued jobs of all the strategies we depend on.

    param pool: pool where the jobs will be queued
    type pool: bsb.services.pool.JobPool
    """
    deps = set(_gutil.ichain(pool.get_submissions_of(strat) for strat in self.get_deps()))
    # Schedule all chunks in 1 job
    pre_chunks = _all_chunks(self.presynaptic.cell_types)
    post_chunks = _all_chunks(self.postsynaptic.cell_types)
    _job = pool.queue_connectivity(self, pre_chunks, post_chunks, deps=deps)


def _raise_na(*args, **kwargs):
    raise NotImplementedError("NotParallel connection strategies have no RoI.")


class HasDependencies:
    """
    Mixin class to mark that this node may depend on other nodes.
    """

    @_abc.abstractmethod
    def get_deps(self):
        pass

    @_abc.abstractmethod
    def __lt__(self, other):
        raise NotImplementedError(f"{type(self).__name__} must implement __lt__.")

    @_abc.abstractmethod
    def __hash__(self):
        raise NotImplementedError(f"{type(self).__name__} must implement __hash__.")

    @classmethod
    def sort_deps(cls, objects):
        """
        Orders a given dictionary of objects by the class's default mechanism and then
        apply the `after` attribute for further restrictions.
        """
        objects = set(objects)
        ordered = []
        sorter = TopologicalSorter(
            {o: set(d for d in o.get_deps() if d in objects) for o in objects}
        )
        sorter.prepare()
        while sorter.is_active():
            node_group = sorter.get_ready()
            ordered.extend(sorted(node_group))
            sorter.done(*node_group)
        return ordered


class NotParallel:
    def __init_subclass__(cls, **kwargs):
        from .connectivity import ConnectionStrategy
        from .placement import PlacementStrategy

        super().__init_subclass__(**kwargs)
        if PlacementStrategy in cls.__mro__:
            cls.queue = _queue_placement
        elif ConnectionStrategy in cls.__mro__:
            cls.queue = _queue_connectivity
            if "get_region_of_interest" not in cls.__dict__:
                cls.get_region_of_interest = _raise_na
        else:
            raise Exception(
                "NotParallel can only be applied to placement or connectivity strategies"
            )


class InvertedRoI:
    """
    This mixin inverts the perspective of the ``get_region_of_interest`` interface and
    lets you find presynaptic regions of interest for a postsynaptic chunk.

    Usage:

    ..code-block:: python

        class MyConnStrat(InvertedRoI, ConnectionStrategy):
          def get_region_of_interest(post_chunk):
            return [pre_chunk1, pre_chunk2]
    """

    def queue(self, pool):
        # Get the queued jobs of all the strategies we depend on.
        deps = set(
            _gutil.ichain(pool.get_submissions_of(strat) for strat in self.get_deps())
        )
        post_types = self.postsynaptic.cell_types
        # Iterate over each chunk that is populated by our postsynaptic cell types.
        to_chunks = set(
            _gutil.ichain(ct.get_placement_set().get_all_chunks() for ct in post_types)
        )
        rois = {
            chunk: roi
            for chunk in to_chunks
            if (roi := self.get_region_of_interest(chunk)) is None or len(roi)
        }
        if not rois:
            warn(
                f"No overlap found between {[post.name for post in post_types]} and "
                f"{[pre.name for pre in self.presynaptic.cell_types]} "
                f"in '{self.name}'."
            )
        for chunk, roi in rois.items():
            pool.queue_connectivity(self, roi, [chunk], deps=deps)


__all__ = ["HasDependencies", "InvertedRoI", "NotParallel"]



"""
Job pooling module.

Jobs derive from the base :class:`.Job` class which can be put on the queue of a
:class:`.JobPool`. In order to submit themselves to the pool Jobs will
:meth:`~.Job.serialize` themselves into a predefined set of variables::

   job.serialize() -> (job_type, f, args, kwargs)

* ``job_type`` should be a string that is a class name defined in this module.
   (e.g. ``"PlacementJob"``)

* ``f`` should be the function object that the job's ``execute`` method should
   execute.

* ``args`` and ``kwargs`` are the args to be passed to that ``f``.

The :meth:`.Job.execute` handler can help interpret ``args`` and ``kwargs``
before running ``f``. The execute handler has access to the scaffold on the MPI
process so one best serializes just the name of some part of the configuration,
rather than trying to pickle the complex objects. For example, the
:class:`.PlacementJob` uses the first ``args`` element to store the
:class:`~bsb.placement.strategy.PlacementStrategy` name and then retrieve it from the
scaffold:

.. code-block:: python

    @staticmethod
    def execute(job_owner, f, args, kwargs):
        placement = job_owner.placement[args[0]]
        indicators = placement.get_indicators()
        return f(placement, *args[1:], indicators, **kwargs)

A job has a couple of display variables that can be set: ``_cname`` for the
class name, ``_name`` for the job name and ``_c`` for the chunk. These are used
to display what the workers are doing during parallel execution. This is an experimental
API and subject to sudden change in the future.
"""

import abc
import concurrent.futures
import contextlib
import functools
import logging
import pickle
import tempfile
import threading
import typing
import warnings
import zlib
from contextlib import ExitStack
from enum import Enum, auto

import numpy as np
from exceptiongroup import ExceptionGroup

from .._util import obj_str_insert
from ..exceptions import (
    JobCancelledError,
    JobPoolContextError,
    JobPoolError,
    JobSchedulingError,
)
from ._util import ErrorModule, MockModule

if typing.TYPE_CHECKING:
    from mpipool import MPIExecutor


class WorkflowError(ExceptionGroup):
    pass


class JobErroredError(Exception):
    def __init__(self, message, error):
        super().__init__(message)
        self.error = error


class JobStatus(Enum):
    # Job has not been queued yet, waiting for dependencies to resolve.
    PENDING = "pending"
    # Job is on the queue.
    QUEUED = "queued"
    # Job is currently running on a worker.
    RUNNING = "running"
    # Job ran successfully.
    SUCCESS = "success"
    # Job failed (an exception was raised).
    FAILED = "failed"
    # Job was cancelled before it started running.
    CANCELLED = "cancelled"
    # Job was killed for some reason.
    ABORTED = "aborted"


class PoolStatus(Enum):
    # Pool has been initialized and jobs can be scheduled.
    SCHEDULING = "scheduling"
    # Pool started execution.
    EXECUTING = "executing"
    # Pool is closing down.
    CLOSING = "closing"


class PoolProgressReason(Enum):
    POOL_STATUS_CHANGE = auto()
    JOB_ADDED = auto()
    JOB_STATUS_CHANGE = auto()
    MAX_TIMEOUT_PING = auto()


class Workflow:
    def __init__(self, phases: list[str]):
        self._phases = phases
        self._phase = 0

    @property
    def phases(self):
        return [*self._phases]

    @property
    def finished(self):
        return self._phase >= len(self._phases)

    @property
    def phase(self):
        if self.finished:
            return "finished"
        else:
            return self._phases[self._phase]

    def next_phase(self):
        self._phase += 1
        return self.phase


class PoolProgress:
    """
    Class used to report pool progression to listeners.
    """

    def __init__(self, pool: "JobPool", reason: PoolProgressReason):
        self._pool = pool
        self._reason = reason

    @property
    def reason(self):
        return self._reason

    @property
    def workflow(self):
        return self._pool.workflow

    @property
    def jobs(self):
        return self._pool.jobs

    @property
    def status(self):
        return self._pool.status


class PoolJobAddedProgress(PoolProgress):
    def __init__(self, pool: "JobPool", job: "Job"):
        super().__init__(pool, PoolProgressReason.JOB_ADDED)
        self._job = job

    @property
    def job(self):
        return self._job


class PoolJobUpdateProgress(PoolProgress):
    def __init__(self, pool: "JobPool", job: "Job", old_status: "JobStatus"):
        super().__init__(pool, PoolProgressReason.JOB_STATUS_CHANGE)
        self._job = job
        self._old_status = old_status

    @property
    def job(self):
        return self._job

    @property
    def old_status(self):
        return self._old_status

    @property
    def status(self):
        return self._job.status


class PoolStatusProgress(PoolProgress):
    def __init__(self, pool: "JobPool", old_status: PoolStatus):
        super().__init__(pool, PoolProgressReason.POOL_STATUS_CHANGE)
        self._old_status = old_status


class _MissingMPIExecutor(ErrorModule):
    pass


class _MPIPoolModule(MockModule):
    @property
    def MPIExecutor(self) -> type["MPIExecutor"]:
        return _MissingMPIExecutor(
            "This is not a public interface. Use `.services.JobPool` instead."
        )

    def enable_serde_logging(self):
        import mpipool

        mpipool.enable_serde_logging()


_MPIPool = _MPIPoolModule("mpipool")


def dispatcher(pool_id, job_args):
    """
    The dispatcher is the function that gets pickled on main, and unpacked "here" on the
    worker. Through class variables on `JobPool` and the given `pool_id` we can find the
    pool and scaffold object, and the job function to run.

    Before running a job, the cache is checked for eventual cached items to free up.
    """
    job_type, args, kwargs = job_args
    # Get the static job execution handler from this module
    handler = globals()[job_type].execute
    # Get the owning scaffold from the JobPool class variables, which act as a registry.
    owner = JobPool.get_owner(pool_id)

    # Check the pool's cache
    pool = JobPool._pools[pool_id]
    required_cache_items = pool._read_required_cache_items()
    # and free any stale cached items
    free_stale_pool_cache(owner, required_cache_items)

    # Execute the job handler.
    return handler(owner, args, kwargs)


class SubmissionContext:
    """
    Context information on who submitted a certain job.
    """

    def __init__(self, submitter, chunks=None, **kwargs):
        self._submitter = submitter
        self._chunks = chunks
        self._context = kwargs

    @property
    def name(self):
        if hasattr(self._submitter, "get_node_name"):
            name = self._submitter.get_node_name()
        else:
            name = str(self._submitter)
        return name

    @property
    def submitter(self):
        return self._submitter

    @property
    def chunks(self):
        from ..storage._chunks import chunklist

        return chunklist(self._chunks) if self._chunks is not None else None

    @property
    def context(self):
        return {**self._context}

    def __getattr__(self, key):
        if key in self._context:
            return self._context[key]
        else:
            return self.__getattribute__(key)


class Job(abc.ABC):
    """
    Dispatches the execution of a function through a JobPool.
    """

    def __init__(
        self,
        pool,
        submission_context: SubmissionContext,
        args,
        kwargs,
        deps=None,
        cache_items=None,
    ):
        self.pool_id = pool.id
        self._comm = pool._comm
        self._args = args
        self._kwargs = kwargs
        self._deps = set(deps or [])
        self._submit_ctx = submission_context
        self._completion_cbs = []
        self._status = JobStatus.PENDING
        self._future: concurrent.futures.Future | None = None
        self._thread: threading.Thread | None = None
        self._res_file = None
        self._error = None
        self._cache_items: list[int] = [] if cache_items is None else cache_items

        for j in self._deps:
            j.on_completion(self._dep_completed)

    @obj_str_insert
    def __str__(self):
        return self.description

    @property
    def name(self):
        return self._submit_ctx.name

    @property
    def description(self):
        descr = self.name
        if self.context:
            descr += " (" + ", ".join(f"{k}={v}" for k, v in self.context.items()) + ")"
        return descr

    @property
    def submitter(self):
        return self._submit_ctx.submitter

    @property
    def context(self):
        return self._submit_ctx.context

    @property
    def status(self):
        return self._status

    @property
    def result(self):
        try:
            with open(self._res_file, "rb") as f:
                return pickle.load(f)
        except Exception:
            raise JobPoolError(f"Result of {self} is not available.") from None

    @property
    def error(self):
        return self._error

    def serialize(self):
        """
        Convert the job to a (de)serializable representation.
        """
        name = self.__class__.__name__
        # First arg is to find the static `execute` method so that we don't have to
        # serialize any of the job objects themselves but can still use different handlers
        # for different job types.
        return (name, self._args, self._kwargs)

    @staticmethod
    @abc.abstractmethod
    def execute(job_owner, args, kwargs):
        """
        Job handler.
        """
        pass

    def run(self, timeout=None):
        """
        Execute the job on the current process, in a thread, and return whether the job is
        still running.
        """
        if self._thread is None:

            def target():
                try:
                    # Execute the static handler
                    result = self.execute(self._pool.owner, self._args, self._kwargs)
                except Exception as e:
                    self._future.set_exception(e)
                else:
                    self._future.set_result(result)

            self._thread = threading.Thread(target=target, daemon=True)
            self._thread.start()
        self._thread.join(timeout=timeout)
        if not self._thread.is_alive():
            self._completed()
            return False
        return True

    def on_completion(self, cb):
        self._completion_cbs.append(cb)

    def set_result(self, value):
        dirname = JobPool.get_tmp_folder(self.pool_id)
        try:
            with tempfile.NamedTemporaryFile(
                prefix=dirname + "/", delete=False, mode="wb"
            ) as fp:
                pickle.dump(value, fp)
                self._res_file = fp.name
        except FileNotFoundError as e:
            self.set_exception(e)
        else:
            self.change_status(JobStatus.SUCCESS)

    def set_exception(self, e: Exception):
        self._error = e
        self.change_status(JobStatus.FAILED)

    def _completed(self):
        if self._status != JobStatus.CANCELLED:
            try:
                result = self._future.result()
            except Exception as e:
                self.set_exception(e)
            else:
                self.set_result(result)
        for cb in self._completion_cbs:
            cb(self)

    def _dep_completed(self, dep):
        # Earlier we registered this callback on the completion of our dependencies.
        # When a dep completes we end up here, and we discard it as a dependency as it has
        # finished. If the dep returns an error remove the job from the pool,
        # since the dependency have failed.
        self._deps.discard(dep)
        if dep._status is not JobStatus.SUCCESS:
            self.cancel("Job killed for dependency failure")
        else:
            # When all our dependencies have been discarded we can queue ourselves.
            # Unless the pool is serial, then the pool itself just runs all jobs in order.
            if not self._deps and self._comm.get_size() > 1:
                # self._pool is set when the pool first tried to enqueue us, but we were
                # still waiting for deps, in the `_enqueue` method below.
                self._enqueue(self._pool)

    def _enqueue(self, pool):
        if not self._deps and self._status is JobStatus.PENDING:
            # Go ahead and submit ourselves to the pool, no dependencies to wait for
            # The dispatcher is run on the remote worker and unpacks the data required
            # to execute the job contents.
            self.change_status(JobStatus.QUEUED)
            self._future = pool._submit(dispatcher, self.pool_id, self.serialize())
        else:
            # We have unfinished dependencies and should wait until we can enqueue
            # ourselves when our dependencies haved all notified us of their completion.
            # Store the reference to the pool though, so later in `_dep_completed` we can
            # call `_enqueue` again ourselves!
            self._pool = pool

    def cancel(self, reason: str | None = None):
        self.change_status(JobStatus.CANCELLED)
        self._error = JobCancelledError() if reason is None else JobCancelledError(reason)
        if self._future and not self._future.cancel():
            warnings.warn(
                f"Could not cancel {self}, the job is already running.", stacklevel=2
            )

    def change_status(self, status: JobStatus):
        old_status = self._status
        self._status = status
        try:
            # Closed pools may have been removed from this map already.
            pool = JobPool._pools[self.pool_id]
        except KeyError:
            pass
        else:
            progress = PoolJobUpdateProgress(pool, self, old_status)
            pool.add_notification(progress)


class PlacementJob(Job):
    """
    Dispatches the execution of a chunk of a placement strategy through a JobPool.
    """

    def __init__(self, pool, strategy, chunk, deps=None):
        args = (strategy.name, chunk)
        context = SubmissionContext(strategy, [chunk])
        cache_items = get_node_cache_items(strategy)
        super().__init__(pool, context, args, {}, deps=deps, cache_items=cache_items)

    @staticmethod
    def execute(job_owner, args, kwargs):
        name, chunk = args
        placement = job_owner.placement[name]
        indicators = placement.get_indicators()
        return placement.place(chunk, indicators, **kwargs)


class ConnectivityJob(Job):
    """
    Dispatches the execution of a chunk of a connectivity strategy through a JobPool.
    """

    def __init__(self, pool, strategy, pre_roi, post_roi, deps=None):
        from ..storage._chunks import chunklist

        args = (strategy.name, pre_roi, post_roi)
        context = SubmissionContext(
            strategy, chunks=chunklist((*(pre_roi or []), *(post_roi or [])))
        )
        cache_items = get_node_cache_items(strategy)
        super().__init__(pool, context, args, {}, deps=deps, cache_items=cache_items)

    @staticmethod
    def execute(job_owner, args, kwargs):
        name = args[0]
        connectivity = job_owner.connectivity[name]
        collections = connectivity._get_connect_args_from_job(*args[1:])
        return connectivity.connect(*collections, **kwargs)


class FunctionJob(Job):
    def __init__(self, pool, f, args, kwargs, deps=None, cache_items=None, **context):
        # Pack the function into the args
        args = (f, args)
        # If no submitter was given, set the function as submitter
        context.setdefault("submitter", f)
        super().__init__(
            pool,
            SubmissionContext(**context),
            args,
            kwargs,
            deps=deps,
            cache_items=cache_items,
        )

    @staticmethod
    def execute(job_owner, args, kwargs):
        # Unpack the function from the args
        f, args = args
        return f(job_owner, *args, **kwargs)


class JobPool:
    _next_pool_id = 0
    _pools = {}
    _pool_owners = {}
    _tmp_folders = {}

    def __init__(self, id, scaffold, fail_fast=False, workflow: "Workflow" = None):
        self._schedulers: list[concurrent.futures.Future] = []
        self.id: int = id
        self._scaffold = scaffold
        self._comm = scaffold._comm
        self._unhandled_errors = []
        self._running_futures: list[concurrent.futures.Future] = []
        self._mpipool: MPIExecutor | None = None
        self._job_queue: list[Job] = []
        self._listeners = []
        self._max_wait = 60
        self._status: PoolStatus = None
        self._progress_notifications: list[PoolProgress] = []
        self._workers_raise_unhandled = False
        self._fail_fast = fail_fast
        self._workflow = workflow
        self._cache_buffer = np.zeros(1000, dtype=np.uint64)
        self._cache_window = self._comm.window(self._cache_buffer)

    def __enter__(self):
        self._context = ExitStack()
        tmp_dirname = self._context.enter_context(tempfile.TemporaryDirectory())

        JobPool._pool_owners[self.id] = self._scaffold
        JobPool._pools[self.id] = self
        JobPool._tmp_folders[self.id] = tmp_dirname
        del self._scaffold

        for listener in self._listeners:
            # Pass if listener is not a context manager
            with contextlib.suppress(TypeError, AttributeError):
                self._context.enter_context(listener)
        self.change_status(PoolStatus.SCHEDULING)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._context.__exit__(exc_type, exc_val, exc_tb)
        # Clean up pool/job references
        self._job_queue = []
        del JobPool._pools[self.id]
        del JobPool._pool_owners[self.id]
        del JobPool._tmp_folders[self.id]
        self.id = None

    def add_listener(self, listener, max_wait=None):
        self._max_wait = min(self._max_wait, max_wait or float("+inf"))
        self._listeners.append(listener)

    @property
    def workflow(self):
        return self._workflow

    @property
    def status(self):
        return self._status

    @property
    def jobs(self) -> list[Job]:
        return [*self._job_queue]

    @property
    def parallel(self):
        return self._comm.get_size() > 1

    @classmethod
    def get_owner(cls, id):
        return cls._pool_owners[id]

    @classmethod
    def get_tmp_folder(cls, id):
        return cls._tmp_folders[id]

    @property
    def owner(self):
        return self.get_owner(self.id)

    def is_main(self):
        return self._comm.get_rank() == 0

    def get_submissions_of(self, submitter):
        return [job for job in self._job_queue if job.submitter is submitter]

    def _put(self, job):
        """
        Puts a job onto our internal queue.
        """
        if self._mpipool and not self._mpipool.open:
            raise JobPoolError("No job pool available for job submission.")
        else:
            self.add_notification(PoolJobAddedProgress(self, job))
            self._job_queue.append(job)
            if self._mpipool:
                # This job was scheduled after the MPIPool was opened, so immediately
                # put it on the MPIPool's queue.
                job._enqueue(self)

    def _submit(self, fn, *args, **kwargs):
        if not self._mpipool or not self._mpipool.open:
            raise JobPoolError("No job pool available for job submission.")
        else:
            future = self._mpipool.submit(fn, *args, **kwargs)
            self._running_futures.append(future)
            return future

    def _schedule(self, future: concurrent.futures.Future, nodes, scheduler):
        _failed_nodes = []
        if not future.set_running_or_notify_cancel():
            return
        try:
            for node in nodes:
                failed_deps = [
                    n for n in getattr(node, "depends_on", []) if n in _failed_nodes
                ]
                if failed_deps:
                    _failed_nodes.append(node)
                    ctx = SubmissionContext(
                        node,
                        error=JobSchedulingError(
                            f"Depends on {failed_deps}, whom failed."
                        ),
                    )
                    self._unhandled_errors.append(ctx)
                    continue
                try:
                    scheduler(node)
                except Exception as e:
                    _failed_nodes.append(node)
                    ctx = SubmissionContext(node, error=e)
                    self._unhandled_errors.append(ctx)
        finally:
            future.set_result(None)

    def schedule(self, nodes, scheduler=None):
        if scheduler is None:

            def scheduler(node):
                node.queue(self)

        future = concurrent.futures.Future()
        self._schedulers.append(future)
        thread = threading.Thread(target=self._schedule, args=(future, nodes, scheduler))
        thread.start()

    @property
    def scheduling(self):
        return any(not f.done() for f in self._schedulers)

    def queue(self, f, args=None, kwargs=None, deps=None, **context):
        job = FunctionJob(self, f, args or [], kwargs or {}, deps, [], **context)
        self._put(job)
        return job

    def queue_placement(self, strategy, chunk, deps=None):
        job = PlacementJob(self, strategy, chunk, deps)
        self._put(job)
        return job

    def queue_connectivity(self, strategy, pre_roi, post_roi, deps=None):
        job = ConnectivityJob(self, strategy, pre_roi, post_roi, deps)
        self._put(job)
        return job

    def execute(self, return_results=False):
        """
        Execute the jobs in the queue.

        In serial execution this runs all the jobs in the queue in First In First Out
        order. In parallel execution this enqueues all jobs into the MPIPool unless they
        have dependencies that need to complete first.
        """

        if self.id is None:
            raise JobPoolContextError("Job pools must use a context manager.")

        if self.parallel:
            self._execute_parallel()
        else:
            self._execute_serial()

        if return_results:
            return {
                job: job.result
                for job in self._job_queue
                if job.status == JobStatus.SUCCESS
            }

    def _execute_parallel(self):
        import bsb.options

        # Enable full mpipool debugging
        if bsb.options.debug_pool:
            _MPIPool.enable_serde_logging()
        # Create the MPI pool
        self._mpipool = _MPIPool.MPIExecutor(
            loglevel=logging.DEBUG if bsb.options.debug_pool else logging.CRITICAL
        )

        if self._mpipool.is_worker():
            # The workers will return out of the pool constructor when they receive
            # the shutdown signal from the master, they return here skipping the
            # master logic.

            # Check if we need to abort our process due to errors etc.
            abort = self._comm.bcast(None)
            if abort:
                raise WorkflowError(
                    "Unhandled exceptions during parallel execution.",
                    [JobPoolError("See main node logs for details.")],
                )

            # Free all cached items
            free_stale_pool_cache(self.owner, set())

            return

        try:
            # Tell the listeners execution is running
            self.change_status(PoolStatus.EXECUTING)
            # Kickstart the workers with the queued jobs
            for job in self._job_queue:
                job._enqueue(self)
            # Add the scheduling futures to the running futures, to await them.
            self._running_futures.extend(self._schedulers)
            # Start tracking cached items
            self._update_cache_window()

            # Keep executing as long as any of the schedulers or jobs aren't done yet.
            while self.scheduling or any(
                job.status == JobStatus.PENDING or job.status == JobStatus.QUEUED
                for job in self._job_queue
            ):
                try:
                    done, not_done = concurrent.futures.wait(
                        self._running_futures,
                        timeout=self._max_wait,
                        return_when="FIRST_COMPLETED",
                    )
                except ValueError:
                    # Sometimes a ValueError is raised here, perhaps because we modify
                    # the list below?
                    continue

                # Complete any jobs that are done
                for job in self._job_queue:
                    if job._future in done:
                        job._completed()

                # If a job finished, update the required cache items
                if len(done):
                    self._update_cache_window()

                # Remove running futures that are done
                for future in done:
                    self._running_futures.remove(future)
                # If nothing finished, post a timeout notification.
                if not len(done):
                    self.ping()
                # Notify all the listeners, and store/raise any unhandled errors
                self.notify()

            # Notify listeners that execution is over
            self.change_status(PoolStatus.CLOSING)
            # Raise any unhandled errors
            self.raise_unhandled()
        except:
            # If any exception (including SystemExit and KeyboardInterrupt) happen on main
            # we should broadcast the abort to all worker nodes.
            self._workers_raise_unhandled = True
            raise
        finally:
            # Shut down our internal pool
            self._mpipool.shutdown(wait=False, cancel_futures=True)
            # Broadcast whether the worker nodes should raise an unhandled error.
            self._comm.bcast(self._workers_raise_unhandled)

    def _execute_serial(self):
        # Wait for jobs to finish scheduling
        while concurrent.futures.wait(
            self._schedulers, timeout=self._max_wait, return_when="FIRST_COMPLETED"
        )[1]:
            self.ping()
            self.notify()
        # Prepare jobs for local execution
        for job in self._job_queue:
            job._future = concurrent.futures.Future()
            job._pool = self
            if job.status != JobStatus.CANCELLED and job.status != JobStatus.ABORTED:
                job._status = JobStatus.QUEUED
            else:
                job._future.cancel()

        self.change_status(PoolStatus.EXECUTING)
        # Just run each job serially
        for job in self._job_queue:
            if not job._future.set_running_or_notify_cancel():
                continue
            job.change_status(JobStatus.RUNNING)
            self.notify()
            while job.run(timeout=self._max_wait):
                self.ping()
                self.notify()
            # After each job, check if any cache items can be freed.
            free_stale_pool_cache(self.owner, self.get_required_cache_items())
            self.notify()
        # Raise any unhandled errors
        self.raise_unhandled()

        self.change_status(PoolStatus.CLOSING)

    def change_status(self, status: PoolStatus):
        old_status = self._status
        self._status = status
        self.add_notification(PoolStatusProgress(self, old_status))
        self.notify()

    def add_notification(self, notification: PoolProgress):
        self._progress_notifications.append(notification)

    def ping(self):
        self.add_notification(PoolProgress(self, PoolProgressReason.MAX_TIMEOUT_PING))

    def notify(self):
        for notification in self._progress_notifications:
            job = getattr(notification, "job", None)
            job_error = getattr(job, "error", None)
            has_error = job_error is not None and type(job_error) is not JobCancelledError
            handled_error = [bool(listener(notification)) for listener in self._listeners]
            if has_error and not any(handled_error):
                self._unhandled_errors.append(job)
        if self._fail_fast:
            self.raise_unhandled()
        self._progress_notifications = []

    def raise_unhandled(self):
        if not self._unhandled_errors:
            return
        errors = []
        # Raise and catch for nicer traceback
        for job in self._unhandled_errors:
            try:
                if isinstance(job, SubmissionContext):
                    raise JobSchedulingError(
                        f"{job.name} failed to schedule its jobs."
                    ) from job.context["error"]
                raise JobErroredError(f"{job} failed", job.error) from job.error
            except (JobErroredError, JobSchedulingError) as e:
                errors.append(e)
        self._unhandled_errors = []
        raise WorkflowError(
            "Your workflow encountered errors.",
            errors,
        )

    def get_required_cache_items(self):
        """
        Returns the list of cache functions for all the jobs in the queue.

        :return: set of cache function name
        :rtype: set[int]
        """
        items = set()
        for job in self._job_queue:
            if (
                job.status == JobStatus.QUEUED
                or job.status == JobStatus.PENDING
                or job.status == JobStatus.RUNNING
            ):
                items.update(job._cache_items)
        return items

    def _update_cache_window(self):
        """
        Checks and updates if the cache buffer should be updated by looking at the job
        statuses in the job queue.

        Only call on main.
        """
        # Create a new cache window buffer
        new_buffer = np.zeros(1000, dtype=int)
        for i, item in enumerate(self.get_required_cache_items()):
            new_buffer[i] = item

        # If there are actual cache requirement differences, lock the window
        # and transfer the buffer
        if np.any(new_buffer != self._cache_buffer):
            self._cache_window.Lock(0)
            self._cache_buffer[:] = new_buffer
            self._cache_window.Unlock(0)

    def _read_required_cache_items(self):
        """
        Locks the cache window and read the still required cache items from rank 0.

        Only call on workers.
        """
        from mpi4py.MPI import UINT64_T

        self._cache_window.Lock(0)
        self._cache_window.Get([self._cache_buffer, UINT64_T], 0)
        self._cache_window.Unlock(0)
        return set(self._cache_buffer)


def get_node_cache_items(node):
    return [
        attr.get_pool_cache_id(node)
        for key in dir(node)
        if hasattr(attr := getattr(node, key), "get_pool_cache_id")
    ]


def free_stale_pool_cache(scaffold, required_cache_items: set[int]):
    for stale_key in set(scaffold._pool_cache.keys()) - required_cache_items:
        # If so, pop them and execute the registered cleanup function.
        scaffold._pool_cache.pop(stale_key)()


def pool_cache(caching_function):
    @functools.cache
    def decorated(self, *args, **kwargs):
        self.scaffold.register_pool_cached_item(
            decorated.get_pool_cache_id(self), cleanup
        )
        return caching_function(self, *args, **kwargs)

    def get_pool_cache_id(node):
        if not hasattr(node, "get_node_name"):
            raise RuntimeError(
                "Pool caching can only be used on methods of @node decorated classes."
            )
        return _cache_hash(f"{node.get_node_name()}.{caching_function.__name__}")

    def cleanup():
        decorated.cache_clear()

    decorated.get_pool_cache_id = get_pool_cache_id

    return decorated


def _cache_hash(string):
    return zlib.crc32(string.encode())



from __future__ import annotations

import contextlib
import itertools
import os
import sys
import time
import typing

import numpy as np

from ._util import obj_str_insert
from .config._config import Configuration
from .connectivity import ConnectionStrategy
from .exceptions import (
    ConfigurationSyncError,
    DatasetNotFoundError,
    InputError,
    MissingActiveConfigError,
    NodeNotFoundError,
    RedoError,
)
from .placement import PlacementStrategy
from .profiling import meter
from .reporting import report
from .services import JobPool
from .services._pool_listeners import NonTTYTerminalListener, TTYTerminalListener
from .services.mpi import MPIService
from .services.pool import Job, Workflow
from .simulation import get_simulation_adapter
from .storage import Storage, open_storage
from .storage._chunks import Chunk

if typing.TYPE_CHECKING:
    from .cell_types import CellType
    from .config._config import NetworkNode as Network
    from .postprocessing import AfterConnectivityHook, AfterPlacementHook
    from .simulation.simulation import Simulation
    from .storage.interfaces import (
        ConnectivitySet,
        FileStore,
        MorphologyRepository,
        PlacementSet,
    )
    from .topology import Partition, Region


@meter()
def from_storage(root, comm=None):
    """
    Load :class:`bsb.core.Scaffold` from a storage object.

    :param root: Root (usually path) pointing to the storage object.
    :param mpi4py.MPI.Comm comm: MPI communicator that shares control
      over the Storage.
    :returns: A network scaffold
    :rtype: :class:`bsb.core.Scaffold`
    """
    return open_storage(root, comm).load()


_cfg_props = (
    "network",
    "regions",
    "partitions",
    "cell_types",
    "placement",
    "after_placement",
    "connectivity",
    "after_connectivity",
    "simulations",
)


def _config_property(name):
    def fget(self):
        return getattr(self.configuration, name)

    def fset(self, value):
        setattr(self.configuration, name, value)

    prop = property(fget)
    return prop.setter(fset)


def _bad_flag(flag: bool):
    return flag is not None and bool(flag) is not flag


class Scaffold:
    """
    This is the main object of the bsb package, it represents a network and puts together
    all the pieces that make up the model description such as the
    :class:`~.config.Configuration` with the technical side like the
    :class:`~.storage.Storage`.
    """

    network: Network
    regions: dict[str, Region]
    partitions: dict[str, Partition]
    cell_types: dict[str, CellType]
    placement: dict[str, PlacementStrategy]
    after_placement: dict[str, AfterPlacementHook]
    connectivity: dict[str, ConnectionStrategy]
    after_connectivity: dict[str, AfterConnectivityHook]
    simulations: dict[str, Simulation]

    def __init__(self, config=None, storage=None, clear=False, comm=None):
        """
        Bootstraps a network object.

        :param config: The configuration to use for this network. If it is omitted the
          :external:ref:`bsb.default configuration <default-config>` is used.
        :type config: :class:`~.config.Configuration`
        :param storage: The storage to use to read and write data for this network. If it
          is omitted the configuration's ``Storage`` node is used to construct one.
        :type storage: :class:`~.storage.Storage`
        :param clear: Start with a new network, clearing any previously stored information
        :type clear: bool
        :param comm: MPI communicator that shares control over the Storage.
        :type comm: mpi4py.MPI.Comm
        :returns: A network object
        :rtype: :class:`~.core.Scaffold`
        """
        self._pool_cache: dict[int, typing.Callable[[], None]] = {}
        self._pool_listeners: list[tuple[typing.Callable[[list[Job]], None], float]] = []
        self._configuration = None
        self._storage = None
        self._comm = MPIService(comm)
        self._bootstrap(config, storage, clear=clear)

    def __contains__(self, component):
        return getattr(component, "scaffold", None) is self

    @obj_str_insert
    def __repr__(self):
        file = os.path.abspath(self.storage.root)
        cells_placed = len(self.cell_types)
        n_types = len(self.connectivity)
        return f"'{file}' with {cells_placed} cell types, and {n_types} connection_types"

    def is_main_process(self) -> bool:
        return not self._comm.get_rank()

    def is_worker_process(self) -> bool:
        return bool(self._comm.get_rank())

    def _bootstrap(self, config, storage, clear=False):
        # No config object given, traverse config cascade.
        if config is None:
            # Try loading the storage's active config
            if storage is not None:
                with contextlib.suppress(MissingActiveConfigError):
                    config = storage.load_active_config()

            # No luck? Try loading the project configuration
            if config is None:
                import bsb.options

                path = bsb.options.config
                if path and os.path.exists(path):
                    with open(path) as f:
                        config = bsb.config.parse_configuration_file(f, path=path)

            # Still nothing? Use the default configuration
            if config is None:
                config = Configuration.default()

        if not storage:
            # No storage given, create one.
            report("Creating storage from config.", level=4)
            storage = Storage(
                config.storage.engine,
                config.storage.root,
                self._comm.get_communicator(),
            )
        else:
            # Override MPI comm of storage to match the scaffold's
            storage._comm = self._comm
        if clear:
            # Storage given, but asked to clear it before use.
            storage.remove()
            storage.create()
        # Synchronize the scaffold, config and storage objects for use together
        self._configuration = config
        # Make sure the storage config node reflects the storage we are using
        config._update_storage_node(storage)
        # Give the scaffold access to the uninitialized storage object (for use during
        # config bootstrapping).
        self._storage = storage
        # First, the scaffold is passed to each config node, and their boot methods called
        self._configuration._bootstrap(self)
        # Then, `storage` is initialized for the scaffold, and `config` is stored (happens
        # inside the `storage` property).
        self.storage = storage

    storage_cfg = _config_property("storage")
    for attr in _cfg_props:
        vars()[attr] = _config_property(attr)

    @property
    def configuration(self) -> Configuration:
        return self._configuration

    @configuration.setter
    def configuration(self, cfg: Configuration):
        self._configuration = cfg
        cfg._update_storage_node(self.storage)
        cfg._bootstrap(self)
        self.storage.store_active_config(cfg)

    def sync_config(self):
        import bsb.config

        cfg = None
        try:
            cfg = self.storage.load_active_config()
        except MissingActiveConfigError:
            import bsb.options

            path = bsb.options.config
        else:
            path = cfg._meta.get("path", None)
        if path and os.path.exists(path):
            with open(path) as f:
                cfg = bsb.config.parse_configuration_file(f, path=path)

        if cfg is None:
            raise ConfigurationSyncError("Could not find a configuration to sync")

        self.configuration = cfg

    @property
    def storage(self) -> Storage:
        return self._storage

    @storage.setter
    def storage(self, storage: Storage):
        self._storage = storage
        storage.init(self)

    @property
    def morphologies(self) -> MorphologyRepository:
        return self.storage.morphologies

    @property
    def files(self) -> FileStore:
        return self.storage.files

    def clear(self):
        """
        Clears the storage.

        This deletes any existing network data!
        """
        self.storage.renew(self)

    def clear_placement(self):
        """
        Clears the placement storage.
        """
        self.storage.clear_placement(self)

    def clear_connectivity(self):
        """
        Clears the connectivity storage.
        """
        self.storage.clear_connectivity()

    def resize(self, x=None, y=None, z=None):
        """
        Updates the topology boundary indicators.

        Use before placement, updates only the abstract topology tree, does not rescale,
        prune or otherwise alter already existing placement data.
        """
        from .topology._layout import box_layout

        if x is not None:
            self.network.x = x
        if y is not None:
            self.network.y = y
        if z is not None:
            self.network.z = z
        self.topology.do_layout(
            box_layout(
                self.network.origin,
                np.array(self.network.origin)
                + [self.network.x, self.network.y, self.network.z],
            )
        )

    @meter()
    def run_placement(self, strategies=None, fail_fast=True, pipelines=True):
        """
        Run placement strategies.
        """
        if pipelines:
            self.run_pipelines()
        if strategies is None:
            strategies = set(self.placement.values())
        strategies = PlacementStrategy.sort_deps(strategies)
        with self.create_job_pool(fail_fast=fail_fast) as pool:
            if pool.is_main():

                def scheduler(strategy):
                    strategy.queue(pool, self.network.chunk_size)

                pool.schedule(strategies, scheduler)
            pool.execute()

    @meter()
    def run_connectivity(self, strategies=None, fail_fast=True, pipelines=True):
        """
        Run connection strategies.
        """
        if pipelines:
            self.run_pipelines()
        if strategies is None:
            strategies = set(self.connectivity.values())
        strategies = ConnectionStrategy.sort_deps(strategies)
        with self.create_job_pool(fail_fast=fail_fast) as pool:
            if pool.is_main():
                pool.schedule(strategies)
            pool.execute()

    @meter()
    def run_placement_strategy(self, strategy):
        """
        Run a single placement strategy.
        """
        self.run_placement([strategy])

    @meter()
    def run_after_placement(self, hooks=None, fail_fast=None, pipelines=True):
        """
        Run after placement hooks.
        """
        if hooks is None:
            hooks = set(self.after_placement.values())
        with self.create_job_pool(fail_fast) as pool:
            if pool.is_main():
                pool.schedule(hooks)
            pool.execute()

    @meter()
    def run_after_connectivity(self, hooks=None, fail_fast=None, pipelines=True):
        """
        Run after placement hooks.
        """
        if hooks is None:
            hooks = set(self.after_connectivity.values())
        with self.create_job_pool(fail_fast) as pool:
            if pool.is_main():
                pool.schedule(hooks)
            pool.execute()

    @meter()
    def compile(
        self,
        skip_placement=False,
        skip_connectivity=False,
        skip_after_placement=False,
        skip_after_connectivity=False,
        only=None,
        skip=None,
        clear=False,
        append=False,
        redo=False,
        force=False,
        fail_fast=True,
    ):
        """
        Run reconstruction steps in the scaffold sequence to obtain a full network.
        """
        existed = self.storage.preexisted

        p_strats = [] if skip_placement else self.get_placement(skip=skip, only=only)
        c_strats = (
            [] if skip_connectivity else self.get_connectivity(skip=skip, only=only)
        )
        todo_list_str = ", ".join(s.name for s in itertools.chain(p_strats, c_strats))
        report(f"Compiling the following strategies: {todo_list_str}", level=2)
        if _bad_flag(clear) or _bad_flag(redo) or _bad_flag(append):
            raise InputError(
                "`clear`, `redo` and `append` are strictly boolean flags. "
                "Pass the strategies to run to the skip/only options instead."
            )
        if sum((bool(clear), bool(redo), bool(append))) > 1:
            raise InputError("`clear`, `redo` and `append` are mutually exclusive.")
        if existed:
            if not (clear or append or redo):
                raise FileExistsError(
                    f"The `{self.storage.format}` storage"
                    + f" at `{self.storage.root}` already exists. Either move/delete it,"
                    + " or pass one of the `clear`, `append` or `redo` arguments"
                    + " to pick what to do with the existing data."
                )
            if clear:
                report("Clearing data", level=2)
                # Clear the placement and connectivity data, but leave any cached files
                # and morphologies intact.
                self.clear()
            elif redo:
                # In order to properly redo things, we clear some placement and connection
                # data, but since multiple placement/connection strategies can contribute
                # to the same sets we might be wiping their data too, and they will need
                # to be cleared and reran as well.
                p_strats, c_strats = self._redo_chain(p_strats, c_strats, skip, force)
            # else:
            #   append mode is luckily simpler, just don't clear anything :)

        phases = ["pipelines"]
        if not skip_placement:
            phases.append("placement")
        if not skip_after_placement:
            phases.append("after_placement")
        if not skip_connectivity:
            phases.append("connectivity")
        if not skip_after_connectivity:
            phases.append("after_connectivity")
        self._workflow = Workflow(phases)
        try:
            self.run_pipelines(fail_fast=fail_fast)
            self._workflow.next_phase()
            if not skip_placement:
                placement_todo = ", ".join(s.name for s in p_strats)
                report(f"Starting placement strategies: {placement_todo}", level=2)
                self.run_placement(p_strats, fail_fast=fail_fast, pipelines=False)
                self._workflow.next_phase()
            if not skip_after_placement:
                self.run_after_placement(pipelines=False, fail_fast=fail_fast)
                self._workflow.next_phase()
            if not skip_connectivity:
                connectivity_todo = ", ".join(s.name for s in c_strats)
                report(f"Starting connectivity strategies: {connectivity_todo}", level=2)
                self.run_connectivity(c_strats, fail_fast=fail_fast, pipelines=False)
                self._workflow.next_phase()
            if not skip_after_connectivity:
                self.run_after_connectivity(pipelines=False)
                self._workflow.next_phase()
        finally:
            # After compilation we should flag the storage as having existed before
            # so that the `clear`, `redo` and `append` flags take effect on a second
            # `compile` pass.
            self.storage._preexisted = True
            del self._workflow

    @meter()
    def run_pipelines(self, fail_fast=True, pipelines=None):
        if pipelines is None:
            pipelines = self.get_dependency_pipelines()
        with self.create_job_pool(fail_fast=fail_fast) as pool:
            if pool.is_main():
                pool.schedule(pipelines)
            pool.execute()

    @meter()
    def run_simulation(self, simulation_name: str):
        """
        Run a simulation starting from the default single-instance adapter.

        :param simulation_name: Name of the simulation in the configuration.
        :type simulation_name: str
        """
        simulation = self.get_simulation(simulation_name)
        adapter = get_simulation_adapter(
            simulation.simulator, comm=self._comm.get_communicator()
        )
        return adapter.simulate(simulation)[0]

    def get_simulation(self, sim_name: str) -> Simulation:
        """
        Retrieve the default single-instance adapter for a simulation.
        """
        if sim_name not in self.simulations:
            simstr = ", ".join(f"'{s}'" for s in self.simulations)
            raise NodeNotFoundError(
                f"Unknown simulation '{sim_name}', choose from: {simstr}"
            )
        return self.configuration.simulations[sim_name]

    def place_cells(
        self,
        cell_type,
        positions,
        morphologies=None,
        rotations=None,
        additional=None,
        chunk=None,
    ):
        """
        Place cells inside the scaffold.

        .. code-block:: python

            # Add one granule cell at position 0, 0, 0
            cell_type = scaffold.get_cell_type("granule_cell")
            scaffold.place_cells(cell_type, cell_type.layer_instance, [[0.0, 0.0, 0.0]])

        :param cell_type: The type of the cells to place.
        :type cell_type: ~bsb.cell_types.CellType
        :param positions: A collection of xyz positions to place the cells on.
        :type positions: Any `np.concatenate` type of shape (N, 3).
        """
        if chunk is None:
            chunk = Chunk([0, 0, 0], self.network.chunk_size)
        if hasattr(chunk, "dimensions") and np.any(np.isnan(chunk.dimensions)):
            chunk.dimensions = self.network.chunk_size
        self.get_placement_set(cell_type).append_data(
            chunk,
            positions=positions,
            morphologies=morphologies,
            rotations=rotations,
            additional=additional,
        )

    def create_entities(self, cell_type, count):
        """
        Create entities in the simulation space.

        Entities are different from cells because they have no positional data and
        don't influence the placement step. They do have a representation in the
        connection and simulation step.

        :param cell_type: The cell type of the entities
        :type cell_type: ~bsb.cell_types.CellType
        :param count: Number of entities to place
        :type count: int
        :todo: Allow `additional` data for entities
        """
        if count == 0:
            return
        ps = self.get_placement_set(cell_type)
        # Append entity data to the default chunk 000
        chunk = Chunk([0, 0, 0], self.network.chunk_size)
        ps.append_entities(chunk, count)

    def get_placement(
        self, cell_types=None, skip=None, only=None
    ) -> list[PlacementStrategy]:
        if cell_types is not None:
            cell_types = [
                self.cell_types[ct] if isinstance(ct, str) else ct for ct in cell_types
            ]
        return [
            val
            for key, val in self.placement.items()
            if (cell_types is None or any(ct in cell_types for ct in val.cell_types))
            and (only is None or key in only)
            and (skip is None or key not in skip)
        ]

    def get_placement_of(self, *cell_types):
        """
        Find all the placement strategies that involve the given cell types.

        :param cell_types: Cell types (or their names) of interest.
        :type cell_types: ~bsb.cell_types.CellType | str
        """
        return self.get_placement(cell_types=cell_types)

    def get_placement_set(
        self, type, chunks=None, labels=None, morphology_labels=None
    ) -> PlacementSet:
        """
        Return a cell type's placement set from the output formatter.

        :param type: Cell type name
        :type type: ~bsb.cell_types.CellType | str
        :param chunks: Optionally load a specific list of chunks.
        :type chunks: list[tuple[float, float, float]]
        :param labels: Labels to filter the placement set by.
        :type labels: list[str]
        :param morphology_labels: Subcellular labels to apply to the morphologies.
        :type morphology_labels: list[str]
        :returns: A placement set
        :rtype: :class:`~.storage.interfaces.PlacementSet`
        """
        if isinstance(type, str):
            type = self.cell_types[type]
        return self.storage.get_placement_set(
            type, chunks=chunks, labels=labels, morphology_labels=morphology_labels
        )

    def get_placement_sets(self) -> list[PlacementSet]:
        """
        Return all the placement sets present in the network.

        :rtype: list[~bsb.storage.interfaces.PlacementSet]
        """
        return [cell_type.get_placement_set() for cell_type in self.cell_types.values()]

    def connect_cells(self, pre_set, post_set, src_locs, dest_locs, name):
        """
        Connect cells from a presynaptic placement set to cells of a postsynaptic
        placement set, and into a connectivity set.
        The description of the hemitype (source or target cell population)
        connection location is stored as a list of 3 ids: the cell index
        (in the placement set), morphology branch index, and the morphology branch
        section index. If no morphology is attached to the hemitype,
        then the morphology indexes can be set to -1.

        :param bsb.storage.interfaces.PlacementSet pre_set: presynaptic placement set
        :param bsb.storage.interfaces.PlacementSet post_set: postsynaptic placement set
        :param list[list[int, int, int]] src_locs: list of the presynaptic
          `connection location`.
        :param list[list[int, int, int]] dest_locs: list of the postsynaptic
          `connection location`.
        :param str name: Name to give to the `ConnectivitySet`
        """
        cs = self.require_connectivity_set(pre_set.cell_type, post_set.cell_type, name)
        cs.connect(pre_set, post_set, src_locs, dest_locs)

    def get_connectivity(
        self, anywhere=None, presynaptic=None, postsynaptic=None, skip=None, only=None
    ) -> list[ConnectivitySet]:
        conntype_filtered = self._connectivity_query(
            any_query=set(self._sanitize_ct(anywhere)),
            pre_query=set(self._sanitize_ct(presynaptic)),
            post_query=set(self._sanitize_ct(postsynaptic)),
        )
        return [
            ct
            for ct in conntype_filtered
            if (only is None or ct.name in only) and (skip is None or ct.name not in skip)
        ]

    def get_connectivity_sets(self) -> list[ConnectivitySet]:
        """
        Return all connectivity sets from the output formatter.

        :returns: All connectivity sets
        :rtype: list[:class:`~.storage.interfaces.ConnectivitySet`]
        """
        return [self._load_cs_types(cs) for cs in self.storage.get_connectivity_sets()]

    def require_connectivity_set(self, pre, post, tag=None) -> ConnectivitySet:
        return self._load_cs_types(
            self.storage.require_connectivity_set(pre, post, tag), pre, post
        )

    def get_connectivity_set(self, tag=None, pre=None, post=None) -> ConnectivitySet:
        """
        Return a connectivity set from its name according to the output formatter. The
        name can be specified directly with tag or with deduced from pre and post if there
        is only one connectivity set matching this pair.

        :param tag: Unique identifier of the connectivity set in the output formatter
        :type tag: str
        :param pre: Presynaptic cell type
        :type pre: ~bsb.cell_types.CellType
        :param post: Postsynaptic cell type
        :type post: ~bsb.cell_types.CellType
        :returns: A connectivity set
        :rtype: :class:`~.storage.interfaces.ConnectivitySet`
        """
        if tag is None:
            try:
                tag = f"{pre.name}_to_{post.name}"
            except Exception:
                raise ValueError(
                    "Supply either `tag` or a valid pre and post cell type."
                ) from None
        return self._load_cs_types(self.storage.get_connectivity_set(tag), pre, post)

    def get_cell_types(self) -> list[CellType]:
        """
        Return a list of all cell types in the network.
        """
        return [*self.configuration.cell_types.values()]

    def merge(self, other, label=None):
        raise NotImplementedError("Revisit: merge CT, PS & CS, done?")

    def _sanitize_ct(self, seq_str_or_none):
        if seq_str_or_none is None:
            return []
        try:
            if isinstance(seq_str_or_none, str):
                return [self.cell_types[seq_str_or_none]]
            return [
                self.cell_types[s] if isinstance(s, str) else s for s in seq_str_or_none
            ]
        except KeyError as e:
            raise NodeNotFoundError(f"Cell type `{e.args[0]}` not found.") from None

    def _connectivity_query(self, any_query=None, pre_query=None, post_query=None):
        # Filter network connection types for any type that satisfies both
        # the presynaptic and postsynaptic query. Empty queries satisfy all
        # types. The presynaptic query is satisfied if the conn type contains
        # any of the queried cell types presynaptically, and same for post.
        # The any query is satisfied if a cell type is found either pre or post.

        if post_query is None:
            post_query = set()
        if pre_query is None:
            pre_query = set()
        if any_query is None:
            any_query = set()

        def partial_query(types, query):
            return not query or any(cell_type in query for cell_type in types)

        def query(conn_type):
            pre_match = partial_query(conn_type.presynaptic.cell_types, pre_query)
            post_match = partial_query(conn_type.postsynaptic.cell_types, post_query)
            any_match = partial_query(
                conn_type.presynaptic.cell_types, any_query
            ) or partial_query(conn_type.postsynaptic.cell_types, any_query)
            return any_match and pre_match and post_match

        types = self.connectivity.values()
        return [*filter(query, types)]

    def _redo_chain(self, p_strats, c_strats, skip, force):
        p_contrib = set(p_strats)
        while True:
            # Get all the placement strategies that effect the current set of CT.
            cell_types_affected = set(
                itertools.chain(*(ps.cell_types for ps in p_contrib))
            )
            contrib = set(self.get_placement(cell_types_affected))
            if contrib:
                contrib |= set(
                    ps
                    for ps in self.get_placement()
                    if np.any(np.isin(ps.depends_on, list(contrib)))
                )
            # Keep repeating until no new contributors are fished up.
            if contrib.issubset(p_contrib):
                break
            # Grow the placement chain
            p_contrib.update(contrib)
        report(
            "Redo-affected placement: " + " ".join(ps.name for ps in p_contrib), level=2
        )

        c_contrib = set(c_strats) | set(
            []
            if len(cell_types_affected) == 0
            else self.get_connectivity(anywhere=cell_types_affected)
        )
        while True:
            contrib = c_contrib | set(
                cs
                for cs in self.get_connectivity()
                if np.any(np.isin(cs.depends_on, list(c_contrib)))
            )
            if contrib.issubset(c_contrib):
                break
            c_contrib.update(contrib)
        report(
            "Redo-affected connectivity: " + " ".join(cs.name for cs in c_contrib),
            level=2,
        )
        # Don't do greedy things without `force`
        # Error if we need to redo things the user asked to skip
        if not force and skip is not None:
            unskipped = [p.name for p in p_contrib if p.name in skip]
            if unskipped:
                chainstr = ", ".join(f"'{s.name}'" for s in (p_strats + c_strats))
                skipstr = ", ".join(f"'{s}'" for s in unskipped)
                raise RedoError(
                    f"Can't skip {skipstr}. Redoing {chainstr} requires to redo them."
                    + f" Omit {skipstr} from `skip` or use `force` (not recommended)."
                )

        for ct in cell_types_affected:
            report(f"Clearing all data of {ct.name}", level=2)
            ct.clear()

        c_sets = set(itertools.chain(*(cs.get_output_names() for cs in c_contrib)))
        for cs in c_sets:
            report(f"Clearing connectivity data of {cs}", level=2)
            try:
                cs = self.get_connectivity_set(cs)
            except DatasetNotFoundError:
                continue
            cs.clear()

        return p_contrib, c_contrib

    def get_dependency_pipelines(self):
        return [*self.configuration.morphologies]

    def get_config_diagram(self):
        from .config import make_configuration_diagram

        return make_configuration_diagram(self.configuration)

    def get_storage_diagram(self):
        dot = f'digraph "{self.configuration.name or "network"}" {{'
        for ps in self.get_placement_sets():
            dot += f'\n  {ps.tag}[label="{ps.tag} ({len(ps)} {ps.cell_type.name})"]'
        for conn in self.get_connectivity_sets():
            dot += f"\n  {conn.pre_type.name} -> {conn.post_type.name}"
            dot += f'[label="{conn.tag} ({len(conn)})"];'

        dot += "\n}\n"
        return dot

    def _load_cs_types(self, cs: ConnectivitySet, pre=None, post=None) -> ConnectivitySet:
        if pre and pre.name != cs.pre_type_name:
            raise ValueError(
                "Given and stored type mismatch:" + f" {pre.name} vs {cs.pre_type_name}"
            )
        if post and post.name != cs.post_type_name:
            raise ValueError(
                "Given and stored type mismatch:" + f" {post.name} vs {cs.post_type_name}"
            )
        try:
            cs.pre_type = self.cell_types[cs.pre_type_name]
            cs.post_type = self.cell_types[cs.post_type_name]
        except KeyError as e:
            raise NodeNotFoundError(
                f"Couldn't load '{cs.tag}' connections, missing cell type '{e.args[0]}'."
            ) from None
        return cs

    def create_job_pool(self, fail_fast=None, quiet=False):
        id_pool = self._comm.bcast(int(time.time()), root=0)
        pool = JobPool(
            id_pool,
            self,
            fail_fast=fail_fast,
            workflow=getattr(self, "_workflow", None),
        )
        try:
            # Check whether stdout is a TTY, and that it is larger than 0x0
            # (e.g. MPI sets it to 0x0 unless an xterm is emulated).
            tty = os.isatty(sys.stdout.fileno()) and sum(os.get_terminal_size())
        except Exception:
            tty = False
        if tty:
            fps = 25
            default_listener = TTYTerminalListener(fps)
            default_max_wait = 1 / fps
        else:
            default_listener = NonTTYTerminalListener()
            default_max_wait = None
        if self._pool_listeners:
            for listener, max_wait in self._pool_listeners:
                pool.add_listener(listener, max_wait=max_wait)
        elif not quiet:
            pool.add_listener(default_listener, max_wait=default_max_wait)
        return pool

    def register_listener(self, listener, max_wait=None):
        self._pool_listeners.append((listener, max_wait))

    def remove_listener(self, listener):
        for i, (l_, _) in enumerate(self._pool_listeners):
            if l_ is listener:
                self._pool_listeners.pop(i)
                break

    def register_pool_cached_item(self, id, cleanup):
        """
        Registers a cleanup function for items cached during a parallel workflow. Internal
        use only.

        :param id: Id of the cached item. Should be unique but identical across MPI nodes
        :param cleanup: A callable that cleans up the cached item.
        """
        if id in self._pool_cache:
            raise RuntimeError(f"Pool cache item '{id}' already exists.")
        self._pool_cache[id] = cleanup


class ReportListener:
    def __init__(self, scaffold, file):
        self.file = file
        self.scaffold = scaffold

    def __call__(self, progress):
        report(
            str(progress.progression)
            + "+"
            + str(progress.duration)
            + "+"
            + str(progress.time),
            token="simulation_progress",
        )


__all__ = ["ReportListener", "Scaffold", "from_storage"]



"""
Plugins module.

Uses ``pkg_resources`` to detect installed plugins and loads them as
categories.
"""

import types
from collections import defaultdict
from importlib.metadata import entry_points
from itertools import chain

import errr

from .exceptions import PluginError


# Before 3.10 `importlib.metadata` was provisional, and didn't have `select` yet.
class _EntryPointsPatch(dict):
    def select(self, *, group=None):
        return self.get(group, [])


class mutdict(dict):
    pass


class mutlist(list):
    pass


def discover(category):
    """
    Discover all plugins for a given category.

    :param category: Plugin category (e.g. ``adapters`` to load all ``bsb.adapters``)
    :type category: str
    :returns: Loaded plugins by name.
    :rtype: dict
    """
    registry = {}
    eps = entry_points()
    if not hasattr(eps, "select"):
        eps = _EntryPointsPatch(eps)

    for entry in chain(eps.select(group="bsb." + category), _unittest_plugins[category]):
        try:
            advert = entry.load()
            if hasattr(advert, "__plugin__"):
                advert = advert.__plugin__
            # Use `types.FunctionType` over `callable` as `callable` might confuse plugin
            # objects that have a `__call__` method with plugin factory functions.
            if isinstance(advert, types.FunctionType):
                advert = advert()
            advert = _decorate_advert(advert, entry)
            registry[entry.name] = advert
        except Exception as e:  # pragma: nocover
            errr.wrap(
                PluginError,
                e,
                entry,
                prepend="Could not instantiate the `%plugin.name%` plugin:\n",
            )

    return registry


def _decorate_advert(advert, entry):
    if type(advert) is list:
        advert = mutlist(advert)
    elif type(advert) is dict:
        advert = mutdict(advert)
    advert._bsb_entry_point = entry
    return advert


# Registry to insert plugins without having to install them,
# intended for testing purposes.
_unittest_plugins = defaultdict(list)

__all__ = ["discover"]
